import json
from pathlib import Path
from typing import Any, Iterator, Union

from typedlogic import Fact
from typedlogic.theories.jsonlog.jsonlog import (
    ListNodeHasMember,
    NodeBooleanValue,
    NodeFloatValue,
    NodeIntValue,
    NodeIsList,
    NodeIsLiteral,
    NodeIsObject,
    NodeNullValue,
    NodeStringValue,
    ObjectNodeLookup,
)


def generate_from_source(source: Union[Path, str, Any]) -> Iterator[Fact]:
    if isinstance(source, Path):
        obj = json.load(source.open())
    elif isinstance(source, str):
        obj = json.loads(source)
    else:
        obj = source
    yield from generate_from_object(obj)

def generate_from_object(obj: Any, jsonpath: str = "/") -> Iterator[Fact]:
    if isinstance(obj, dict):
        yield NodeIsObject(jsonpath)
        for k, v in obj.items():
            child = jsonpath + k + "/"
            yield ObjectNodeLookup(jsonpath, k, child)
            yield from generate_from_object(v, child)
    elif isinstance(obj, list):
        yield NodeIsList(jsonpath)
        for i, v in enumerate(obj):
            child = jsonpath + f"[{i}]"
            yield ListNodeHasMember(jsonpath, i, child)
            yield from generate_from_object(v, child)
    else:
        yield NodeIsLiteral(jsonpath)
        if obj is None:
            yield NodeNullValue(jsonpath)
        elif isinstance(obj, str):
            yield NodeStringValue(jsonpath, obj)
        elif isinstance(obj, bool):
            yield NodeBooleanValue(jsonpath, obj)
        elif isinstance(obj, int):
            yield NodeIntValue(jsonpath, obj)
        elif isinstance(obj, float):
            yield NodeFloatValue(jsonpath, obj)
            # TODO: other numbers
        else:
            raise ValueError(f"Unexpected object type: {type(obj)}")





from dataclasses import dataclass
from typing import ClassVar, Optional, Union

from typedlogic import Theory
from typedlogic.compiler import Compiler, ModelSyntax
from typedlogic.transformations import as_fol


@dataclass
class FOLCompiler(Compiler):

    default_suffix: ClassVar[str] = "fol"

    def compile(self, theory: Theory, syntax: Optional[Union[str, ModelSyntax]] = None, **kwargs) -> str:
        lines = []
        for s in theory.sentences:
            lines.append(as_fol(s))
        return "\n".join(lines)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import TextIO, Union, Any, ClassVar

from typedlogic import Theory


class Parser(ABC):
    """
    A parser is a class that can parse a source and return a Theory object.

    You can use the registry `get_parser` method to get a parser for a particular syntax:

        >>> from typedlogic.registry import get_parser
        >>> parser = get_parser("yaml")


    """

    default_suffix: ClassVar[str] = "txt"

    def parse_file(self, source: Union[Path, str, TextIO], **kwargs) -> Theory:
        if isinstance(source, str):
            source = Path(source)
        return self.parse(source, **kwargs)

    @abstractmethod
    def parse(self, source: Union[Path, str, TextIO], **kwargs) -> Theory:
        """
        Parse a source and return a Theory object.

        :param source: A path to a file, a string representation of the source, or a file-like object.
        :param kwargs:
        :return:
        """
        pass

    def translate(self, source: Any, **kwargs) -> Theory:
        """
        Translate a source object into a Theory object.

        The type of the source object may be constrained by individual subclasses

        :param source:
        :param kwargs:
        :return:
        """
        raise NotImplementedError("Translation not supported by this parser")


import json
from dataclasses import dataclass
from typing import ClassVar, Optional, Union

from typedlogic import Sentence, Theory
from typedlogic.compiler import Compiler, ModelSyntax
from typedlogic.datamodel import SExpression, as_sexpr


def _render(s: SExpression, position=0, depth=0) -> str:
    rendered = ""
    if position > 0:
        rendered += "\n"
        for _ in range(depth):
            rendered += "  "
    if isinstance(s, list):
        return rendered + f"({' '.join([_render(x, i, depth+1) for i, x in enumerate(s)])})"
    if isinstance(s, dict):
        raise ValueError(f"Cannot render dict: {s}")
    if position == 0:
        return str(s)
    return json.dumps(s)

@dataclass
class SExprCompiler(Compiler):

    default_suffix: ClassVar[str] = "sexpr"

    def compile(self, theory: Theory, syntax: Optional[Union[str, ModelSyntax]] = None, **kwargs) -> str:
        """
        Compile a Theory object into S-Expressions.

        Example:
        -------
            >>> from typedlogic import *
            >>> theory = Theory()
            >>> x = Variable("x")
            >>> theory.predicate_definitions = [PredicateDefinition("P", {"x": "str"}),
            ...              PredicateDefinition("Q", {"x": "str"})]
            >>> s = Implies(Term("P", x), Term("Q", x))
            >>> theory.add(sentence=s)
            >>> compiler = SExprCompiler()
            >>> print(compiler.compile(theory))
            (Theory
              (name null)
              (constants
                (dict
                  ()))
              (type_definitions
                (dict
                  ()))
              (predicate_definitions
                ((PredicateDefinition
                    (predicate "P")
                    (arguments
                      (dict
                        ((x "str"))))
                    (description null)
                    (metadata null)
                    (parents null)
                    (python_class null))
                  (PredicateDefinition
                    (predicate "Q")
                    (arguments
                      (dict
                        ((x "str"))))
                    (description null)
                    (metadata null)
                    (parents null)
                    (python_class null))))
              (sentence_groups
                ((SentenceGroup
                    (name "Sentences")
                    (group_type null)
                    (docstring null)
                    (sentences
                      ((Implies
                          (P
                            (Variable "x"))
                          (Q
                            (Variable "x")))))
                    (_annotations null))))
              (ground_terms
                ())
              (_annotations null)
              (source_module_name null))

        :param theory:
        :param syntax:
        :param kwargs:
        :return:

        """
        sexpr = as_sexpr(theory)
        return _render(sexpr)

    def compile_sentence(self, sentence: Sentence, syntax: Optional[Union[str, ModelSyntax]] = None, **kwargs) -> str:
        sexpr = as_sexpr(sentence)
        return _render(sexpr)



import importlib
import inspect
import logging
import os
from pathlib import Path
from typing import Dict, Type, Union, Tuple

from attr import dataclass

import typedlogic
from typedlogic.compiler import Compiler
from typedlogic.parser import Parser
from typedlogic.solver import Solver

logger = logging.getLogger(__name__)

SUFFIXES = {
    Compiler: "compiler",
    Parser: "parser",
    Solver: "solver"
}

Extendable = Union[Compiler, Parser, Solver]


@dataclass
class Registry:
    implementation_classes: Dict[Type, Dict[str, Type]] = {}

    def register(self, name: str, category: Type, impl_class: Type[Extendable]):
        if category not in self.implementation_classes:
            self.implementation_classes[category] = {}
        self.implementation_classes[category][name] = impl_class

    def get_implementation_class(self, category: Type, name: str) -> Type[Extendable]:
        if category not in self.implementation_classes:
            raise ValueError(f"Unknown category: {category}\n" 
                             f"Known categories: {list(self.implementation_classes.keys())}")
        if name not in self.implementation_classes[category]:
            raise ValueError(f"Unknown handle: {name}\n" 
                             f"Known implementations: {list(self.implementation_classes[category].keys())}")
        return self.implementation_classes[category][name]

    def create_instance(self, category: Type, handle: str, **kwargs) -> Extendable:
        implementation_class = self.get_implementation_class(category, handle)
        return implementation_class()

    @classmethod
    def load_implementations(cls, package_path: str):
        registry = cls()
        this_path = importlib.import_module(package_path).__file__
        if this_path is None:
            raise ValueError(f"Error: package {package_path} not found")
        package_dir = os.path.dirname(Path(this_path))
        # traverse the directory recursively
        for root, dirs, files in os.walk(package_dir):
            for filename in files:
                filepath = os.path.join(root, filename)
                if filepath.endswith(".py") and not filepath.startswith("__"):
                    base = filepath.replace(package_dir, "").replace("/", ".")[1:-3]
                    module_name = f"{package_path}.{base}"
                    try:
                        module = importlib.import_module(module_name)
                        for name, obj in inspect.getmembers(module):
                            if inspect.isclass(obj):
                                for t, suffix_name in SUFFIXES.items():
                                    if name.lower().endswith(suffix_name) and issubclass(obj, t):
                                        handle = name.lower().replace(suffix_name, "")
                                        registry.register(handle, t, obj)
                    except ImportError as e:
                        logger.info(f"Error importing {module_name}: {e} - assuming not installed")
        return registry


registry = Registry.load_implementations("typedlogic")

def get_compiler(handle: str, **kwargs) -> Compiler:
    """
    Get a compiler

    >>> from typedlogic.registry import get_compiler
    >>> z3_compiler = get_compiler('z3')
    >>> type(z3_compiler)
    <class 'typedlogic.integrations.solvers.z3.z3_compiler.Z3Compiler'>

    :param handle:
    :param kwargs:
    :return:
    """
    return registry.create_instance(Compiler, handle, **kwargs)


def get_solver(handle: str, **kwargs) -> Solver:
    """
    Get a solver

    >>> from typedlogic.registry import get_solver
    >>> z3_solver = get_solver('z3')
    >>> type(z3_solver)
    <class 'typedlogic.integrations.solvers.z3.z3_solver.Z3Solver'>

    :param handle:
    :param kwargs:
    :return:
    """
    return registry.create_instance(Solver, handle, **kwargs)


def get_parser(handle: str, **kwargs) -> Parser:
    """
    Get a parser

    >>> from typedlogic.registry import get_parser
    >>> python_parser = get_parser('python')
    >>> type(python_parser)
    <class 'typedlogic.parsers.pyparser.python_parser.PythonParser'>
    """
    return registry.create_instance(Parser, handle, **kwargs)



import logging
import os
import sys
from io import TextIOWrapper
from pathlib import Path
from types import ModuleType
from typing import Optional, TextIO, Union

from typedlogic import Theory
from typedlogic.parser import Parser
from typedlogic.parsers.pyparser.introspection import (
    get_module_predicate_definitions,
    get_module_sentence_groups,
    translate_module_to_theory,
)

logger = logging.getLogger(__name__)

def compile_python(python_txt: str, name: Optional[str] = None, package_path: Optional[str] = None) -> ModuleType:
    """
    Compile a Python module from a string

    :param python_txt:
    :param package_path:
    :return:
    """
    if name is None:
        if package_path:
            name = os.path.basename(package_path).split('.')[0]
        else:
            name = 'test'
    spec = compile(python_txt, name, 'exec')
    module = ModuleType(name)
    if package_path:
        package_path_abs = os.path.join(os.getcwd(), package_path)
        # We have to calculate the path to expected path relative to the current working directory
        for path in sys.path:
            if package_path.startswith(path):
                path_from_tests_parent = os.path.relpath(package_path, path)
                break
            if package_path_abs.startswith(path):
                path_from_tests_parent = os.path.relpath(package_path_abs, path)
                break
        else:
            logger.warning(f"There is no established path to {package_path} - compile_python may or may not work")
            path_from_tests_parent = os.path.relpath(package_path, os.path.join(os.getcwd(), '..'))
        module.__package__ = os.path.dirname(os.path.relpath(path_from_tests_parent, os.getcwd())).replace(os.path.sep, '.')
    sys.modules[module.__name__] = module
    exec(spec, module.__dict__)
    return module


class PythonParser(Parser):
    """
    A parser for Python modules that contain axioms.

        >>> parser = PythonParser()
        >>> theory = parser.parse(Path("tests/theorems/mortals.py"))
        >>> assert isinstance(theory, Theory)
        >>> theory.name
        'mortals'
    """

    def transform(self, source: ModuleType, **kwargs) -> Theory:
        """
        Transform a Python module into a Theory

        :param source:
        :param kwargs:
        :return:
        """
        return translate_module_to_theory(source)

    def parse(self, source: Union[Path, str, TextIO, ModuleType], file_name: Optional[str] = None, **kwargs) -> Theory:
        if isinstance(source, ModuleType):
            return translate_module_to_theory(source)
        if isinstance(source, Path):
            with source.open() as f:
                return self.parse(f, file_name=str(source), **kwargs)
        if isinstance(source, str):
            module = compile_python(source, name=None, package_path=file_name)
            sgs = get_module_sentence_groups(source)
            pds = get_module_predicate_definitions(module)
            # get the python module name
            return Theory(
                name=module.__name__,
                predicate_definitions=list(pds.values()),
                sentence_groups=sgs,
                source_module_name=module.__name__,
            )
        if isinstance(source, (TextIOWrapper, TextIO)):
            lines = "\n".join(source.readlines())
            return self.parse(lines, file_name=file_name, **kwargs)
        raise ValueError(f"Unsupported source type: {type(source)}")



from io import TextIOWrapper
from pathlib import Path
from typing import TextIO, Union

import yaml

from typedlogic import Theory
from typedlogic.datamodel import from_object
from typedlogic.parser import Parser


class YAMLParser(Parser):
    """
    A parser for YAML files.
    """

    def parse(self, source: Union[Path, str, TextIO], **kwargs) -> Theory:
        if isinstance(source, Path):
            source = source.open()
        if not isinstance(source, (str, TextIOWrapper)):
            raise ValueError(f"Invalid source type: {type(source)}")
        obj = yaml.safe_load(source)
        return from_object(obj)



from abc import ABC
from dataclasses import dataclass, field
from typing import Collection, Type, Tuple, Optional, ClassVar, Set


class Profile(ABC):
    """A profile of FOL"""

    disjoint_profiles: ClassVar[Set[Type["Profile"]]] = set()

    def impl(self, profile: Type["Profile"]) -> bool:
        """True if self is an implementation of profile.

        Operates under NAF: if it cannot be proven that self is an implementation of profile, it returns None.

        >>> p = WellFoundedSemantics()
        >>> p.impl(WellFoundedSemantics)
        True
        >>> p.impl(ClosedWorld)
        True
        >>> p.impl(OpenWorld)
        False

        """
        return isinstance(self, profile)

    def not_impl(self, profile: Type["Profile"]) -> Optional[bool]:
        """True if self is provably not an implementation of profile

        >>> p = WellFoundedSemantics()
        >>> p.not_impl(WellFoundedSemantics)
        False
        >>> p.impl(ClosedWorld)
        True
        >>> p.not_impl(ClosedWorld)
        False
        >>> p.not_impl(OpenWorld)
        True
        >>> assert p.not_impl(SortedLogic) is None

        """
        if self.impl(profile):
            return False
        disjoints = profile.disjoint_profiles
        if disjoints and any(isinstance(self, p) for p in disjoints):
            return True
        this_disjoints = type(self).disjoint_profiles
        if this_disjoints and any(issubclass(profile, p) for p in this_disjoints):
            return True
        return None


@dataclass
class MixedProfile(Profile):
    """A profile that mixes multiple profiles.md

    >>> p = MixedProfile(Unrestricted(), SortedLogic())
    >>> p.impl(Unrestricted)
    True
    >>> p.impl(SortedLogic)
    True

    """
    profiles: Tuple = field(default_factory=tuple)

    def __init__(self, *profiles: Profile):
        self.profiles = profiles

    def impl(self, profile: Type["Profile"]) -> bool:
        return isinstance(self, profile) or any(p.impl(profile) for p in self.profiles)

    def not_impl(self, profile: Type["Profile"]) -> Optional[bool]:
        provably_not_impl = any(p.not_impl(profile) for p in self.profiles)
        provable_impl = any(p.impl(profile) for p in self.profiles)
        if provable_impl and provably_not_impl:
            raise ValueError(f"Inconsistent profiles.md: {self} is both an implementation and not an implementation of {profile}")
        return provably_not_impl

@dataclass
class ExcludedProfile(Profile):
    """A profile that excludes other profiles.md

    >>> p = ExcludedProfile(WellFoundedSemantics())
    >>> p.impl(WellFoundedSemantics)
    False
    >>> p.not_impl(WellFoundedSemantics)
    True

    >>> p = MixedProfile(Unrestricted(), SortedLogic(), ExcludedProfile(PropositionalLogic()))
    >>> p.impl(Unrestricted)
    True
    >>> p.impl(PropositionalLogic)
    False
    >>> p.not_impl(PropositionalLogic)
    True

    """
    profile: Profile

    def impl(self, profile: Type["Profile"]) -> bool:
        """True if self is an implementation of profile"""
        if self.profile.impl(profile):
            return False
        return super().impl(profile)

    def not_impl(self, profile: Type["Profile"]) -> Optional[bool]:
        """True if self is not an implementation of profile"""
        if isinstance(self.profile, profile):
            return True
        return super().not_impl(profile)

class ComputationalProfile(Profile, ABC):
    """A profile that describes the computational properties of a logic"""

class Decidable(ComputationalProfile):
    """The logic is decidable"""

class Undecidable(ComputationalProfile):
    """The logic is undecidable"""

class LogicalFeature(Profile, ABC):
    """A profile grouping corresponding to a feature of logic"""

class NegationLogic(LogicalFeature):
    """The logic has negation"""

class DisjunctionLogic(LogicalFeature):
    """The logic has disjunction"""

class ConjunctionLogic(LogicalFeature):
    """The logic has conjunction"""

class AllowsComparisonTerms(LogicalFeature):
    """The logic allows comparison terms"""

class ModelMultiplicitySemantics(Profile, ABC):
    """A profile grouping corresponding to a model semantics"""

class SingleModelSemantics(ModelMultiplicitySemantics):
    """The logic has a single model"""

class MultipleModelSemantics(ModelMultiplicitySemantics):
    """The logic has multiple models"""

class Assumption(Profile, ABC):
    """Category of assumptions"""

class OpenWorld(Assumption):
    """The assumption that what is not known to be true is false"""

class ClosedWorld(Assumption):
    """The assumption that what is not known to be true is unknown"""
    disjoint_profiles = {OpenWorld}

class WellFoundedSemantics(ClosedWorld):
    """The assumption that the world is well-founded"""

class ClassicPrologNegationAsFailure(ClosedWorld):
    """The assumption that negation is failure"""

class TypeSystem(Profile, ABC):
    """Category of type systems"""

class UnsortedLogic(TypeSystem):
    """The logic has no type system"""

class SortedLogic(TypeSystem):
    """The logic has a type system"""

class LogicalFamily(Profile, ABC):
    """A profile grouping corresponding to a family of logics"""

class Classical(LogicalFamily):
    """The classical logic family"""

class NonClassical(LogicalFamily):
    """The non-classical logic family"""

class Modal(LogicalFamily):
    """The modal logic family"""

class Temporal(LogicalFamily):
    """The temporal logic family"""

class Paraconsistent(LogicalFamily):
    """The paraconsistent logic family"""

class Intuitionistic(LogicalFamily):
    """The intuitionistic logic family"""

class Probabilistic(LogicalFamily):
    """The probabilistic logic family"""

class LogicalSubset(Profile, ABC):
    """A profile grouping corresponding to a subset of logic"""

class Unrestricted(LogicalSubset, Undecidable):
    """The an unrestricted logic"""

class ClassicDatalog(LogicalSubset, ClosedWorld, SingleModelSemantics):
    """A subset of logic that corresponds to Datalog"""

class DisjunctiveDatalog(LogicalSubset, ClosedWorld, MultipleModelSemantics):
    """A subset of logic that corresponds to Datalog"""

class DescriptionLogic(LogicalSubset, Decidable):
    """A subset of logic that corresponds to Description Logic"""

class OWLProfile(LogicalSubset, ABC):
    """The OWL 2 DL subset of Description Logic.

    Note al OWL Profile is not necessarily a DL profile
    """

class OWL2DL(DescriptionLogic, OWLProfile):
    """The OWL 2 DL subset of Description Logic"""

class ModalLogic(LogicalSubset):
    """A subset of logic that corresponds to Modal Logic"""

class OrderOfLogic(Profile, ABC):
    """Category of orderings"""

class PropositionalLogic(OrderOfLogic):
    """The assumption that all predicates are propositional"""

class FirstOrder(OrderOfLogic):
    """The assumption that all predicates are first-order"""

class HigherOrder(OrderOfLogic):
    """The assumption that predicates may be higher-order"""

class NamedLogic(Profile, ABC):
    """A named logic"""

class AnswerSetProgramming(NamedLogic, WellFoundedSemantics, DisjunctiveDatalog):
    """The Answer Set Programming logic"""

class ReasoningParadigm(Profile, ABC):
    """A profile grouping corresponding to a reasoning paradigm"""

class Inductive(ReasoningParadigm):
    """The inductive reasoning paradigm"""

class Deductive(ReasoningParadigm):
    """The deductive reasoning paradigm"""

class Abductive(ReasoningParadigm):
    """The abductive reasoning paradigm"""

class Monotonic(ReasoningParadigm):
    """The monotonic reasoning paradigm"""

class NonMonotonic(ReasoningParadigm):
    """The non-monotonic reasoning paradigm"""

class ConstraintSolver(ReasoningParadigm):
    """The constraint solver reasoning paradigm"""


class UnspecifiedProfile(Profile):
    """An unspecified profile"""


"""
Decorators for marking functions as axioms and goals.

Example:

```python
    from dataclasses import dataclass
    from typedlogic import Fact
    from typedlogic.decorators import axiom, goal

    @dataclass
    class Dog(Fact):
        unique_name: str

    @dataclass
    class Cat(Fact):
        unique_name: str

    @axiom
    def disjointness(n: str):
        '''Nothing os both a dog and a cat'''
        assert not(Dog(n) and Cat(n))

    @goal
    def unit_test1():
        '''unit test: if Violet is a cat, then it must be provable that Violet is not a dog```
        if Cat('Violet'):
           assert not Dog('Violet')
```

Note: when axioms are expressed directly in python programs, it is possible to use
logical connectives such as `and`, `or`, as well as `if...then` to express implication.

When working directly with objects in the datamodel, it's necessary to use symbols
such as `&` and `~`.

"""
from dataclasses import dataclass, fields
from functools import wraps
from typing import Callable

from typedlogic import FactMixin

AXIOM_REGISTRY = []


def axiom(func: Callable) -> Callable:
    """
    Decorator to mark a function as an axiom.
    
    The marked function is not intended to be executed in a standard python environment.

    The arguments to the function are treated as universally quantified variables
    
    Example usage:
    
        from dataclasses import dataclass
        from typedlogic import Fact
        from typedlogic.decorators import axiom, goal
    
        @dataclass
        class Dog(Fact):
            unique_name: str
    
        @dataclass
        class Cat(Fact):
            unique_name: str
    
        @axiom
        def disjointness(n: str):
            '''Forall n: Dog(n) => not Cat(n)'''
            assert not(Dog(n) and Cat(n))

    The arguments of the wrapped functions are treated as universal quantifiers (Forall x: ...)

    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    AXIOM_REGISTRY.append(func)
    return wrapper


def goal(func: Callable) -> Callable:
    """
    Decorator to mark a function as a goal.

    The `prove_goals` function in a Solver object will attempt to prove the goal.

    The function is not intended to be called directly, but rather to be interpreted
    using logical semantics.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    # AXIOM_REGISTRY.append(func)
    return wrapper


def predicate(cls=None, *, exclude_from_hash=None, **kwargs):
    if exclude_from_hash is None:
        exclude_from_hash = []
    def wrapper(cls):

        if not issubclass(cls, FactMixin):
            cls = type(cls.__name__, (cls, FactMixin), dict(cls.__dict__))

        # Apply the dataclass decorator with eq=True and unsafe_hash=True
        cls = dataclass(eq=True, frozen=True, **kwargs)(cls)

        # Store original __hash__ and __eq__ methods
        original_hash = cls.__hash__
        original_eq = cls.__eq__

        def custom_hash(self):
            return hash(tuple(
                getattr(self, f.name) for f in fields(self)
                if f.name not in exclude_from_hash and f.compare
            ))

        def custom_eq(self, other):
            if not isinstance(other, self.__class__):
                return NotImplemented
            return all(
                getattr(self, f.name) == getattr(other, f.name)
                for f in fields(self)
                if f.name not in exclude_from_hash and f.compare
            )

        cls.__hash__ = custom_hash
        cls.__eq__ = custom_eq

        original_init = cls.__init__

        @wraps(original_init)
        def custom_init(*args, **kwargs):
            instance = original_init(*args, **kwargs)
            print(f"Predicate class created: {instance}")
            return instance

        cls.__init__ = custom_init

        return cls

    return wrapper if cls is None else wrapper(cls)


from dataclasses import dataclass
from typing import ClassVar, List, Optional, Union

from typedlogic import Theory
from typedlogic.compiler import Compiler, ModelSyntax
from typedlogic.datamodel import Sentence, SentenceGroupType
from typedlogic.transformations import as_prover9


@dataclass
class Prover9Compiler(Compiler):

    default_suffix: ClassVar[str] = "prover9"

    def compile(self, theory: Theory, syntax: Optional[Union[str, ModelSyntax]] = None, goals: Optional[List[Sentence]] = None, **kwargs) -> str:
        lines = [f"% Problem: {theory.name}"]
        # Assumptions (axioms)
        lines.append("formulas(assumptions).")
        for sg in theory.sentence_groups:
            if sg.group_type != SentenceGroupType.GOAL:
                for sentence in sg.sentences or []:
                    lines.append(f"    {as_prover9(sentence)}.")
        lines.append("end_of_list.")
        lines.append("")

        # Goals (conjecture)

        lines.append("formulas(goals).")
        for sg in theory.sentence_groups:
            if sg.group_type == SentenceGroupType.GOAL:
                for sentence in sg.sentences or []:
                    lines.append(f"    {as_prover9(sentence)}.")
        if goals:
            for sentence in goals:
                lines.append(f"    {as_prover9(sentence)}.")
        lines.append("end_of_list.")
        return "\n".join(lines)



import csv
import logging
import subprocess
import tempfile
from dataclasses import dataclass, field
from pathlib import Path
from typing import ClassVar, Iterator

from typedlogic.integrations.solvers.souffle.souffle_compiler import SouffleCompiler
from typedlogic.profiles import (
    AllowsComparisonTerms,
    ClassicDatalog,
    MixedProfile,
    Profile,
    SingleModelSemantics,
    SortedLogic,
)
from typedlogic.solver import Model, Solution, Solver
from typedlogic.utils.term_maker import make_terms

logger = logging.getLogger(__name__)

@dataclass
class SouffleSolver(Solver):
    """
    A solver that uses Soufflé.

    Soufflé is a logic programming language inspired by Datalog. For more details,
    see [The Souffle site](https://souffle-lang.github.io/).

    Note that in order to use this integration, you need to [install Souffle](https://souffle-lang.github.io/install)
    and have it on your path.

        >>> from typedlogic.integrations.frameworks.pydantic import FactBaseModel
        >>> class AncestorOf(FactBaseModel):
        ...     ancestor: str
        ...     descendant: str
        >>> solver = SouffleSolver()
        >>> from typedlogic import SentenceGroup, PredicateDefinition
        >>> solver.add_predicate_definition(PredicateDefinition(predicate="AncestorOf", arguments={'ancestor': str, 'descendant': str}))
        >>> solver.add_fact(AncestorOf(ancestor='p1', descendant='p1a'))
        >>> solver.add_fact(AncestorOf(ancestor='p1a', descendant='p1aa'))
        >>> aa = SentenceGroup(name="transitivity-of-ancestor-of")
        >>> solver.add_sentence_group(aa)
        >>> soln = solver.check()

    This solver does not implement the open-world assumption.

        >>> from typedlogic.profiles import OpenWorld
        >>> solver.profile.impl(OpenWorld)
        False

    """

    exec_name: str = field(default="souffle")
    profile: ClassVar[Profile] = MixedProfile(ClassicDatalog(), SortedLogic(), AllowsComparisonTerms(), SingleModelSemantics())

    def models(self) -> Iterator[Model]:
        compiler = SouffleCompiler()
        program = compiler.compile(self.base_theory)
        pdmap = {}

        facts = []
        with tempfile.TemporaryDirectory() as temp_dir:
            # Create output directives for each predicate
            output_files = {}
            input_files = {}
            for pd in self.base_theory.predicate_definitions:
                pred = pd.predicate
                pdmap[pred] = pd
                output_file = Path(temp_dir) / f"{pred}.csv"
                output_files[pred] = str(output_file)
                program += f'\n.output {pred}(IO=file, filename="{output_file}")\n'
                input_file = Path(temp_dir) / f"{pred}__in.csv"
                input_files[pred] = str(input_file)
                program += f'\n.input {pred}(IO=file, filename="{input_file}")\n'
                with open(input_file, 'w', encoding="utf-8") as csvfile:
                    writer = csv.writer(csvfile, delimiter="\t")
                    for term in self.base_theory.ground_terms:
                        if term.predicate == pred:
                            writer.writerow(term.bindings.values())

            with tempfile.NamedTemporaryFile(suffix=".dl", mode="w") as fp:
                fp.write(program)
                fp.flush()
                res = subprocess.run([self.exec_name, fp.name], capture_output=True)
                if res.stderr:
                    msg = res.stderr.decode()
                    import re
                    if re.match(r".*Variable (\S+) only occurs once.*", msg):
                        logger.info(msg)
                    else:
                        logger.error(msg)


            for pred, filename in output_files.items():
                if not Path(filename).exists():
                    continue
                pd = pdmap[pred]
                rows = []
                with open(filename, 'r') as csvfile:
                    reader = csv.reader(csvfile, delimiter="\t")
                    for row in reader:
                        rows.append(row)
                facts.extend(make_terms(rows, pd))

        model = Model(
            source_object=self,
            ground_terms=facts
        )
        yield model


    def check(self) -> Solution:
        return Solution(satisfiable=None)

    def dump(self) -> str:
        compiler = SouffleCompiler()
        return compiler.compile(self.base_theory)



from enum import Enum
from typing import Dict, Any, Optional, Union


class BaseType(Enum):
    """
    Enum for base types in the type system.
    """
    STR = "str"
    INT = "int"
    FLOAT = "float"
    BOOL = "bool"
    LIST = "list"
    DICT = "dict"
    TUPLE = "tuple"
    SET = "set"

# Mapping dictionaries
python_type_mapping: Dict[BaseType, Any] = {
    BaseType.STR: str,
    BaseType.INT: int,
    BaseType.FLOAT: float,
    BaseType.BOOL: bool,
    BaseType.LIST: list,
    BaseType.DICT: dict,
    BaseType.TUPLE: tuple,
    BaseType.SET: set,
}

json_schema_mapping: Dict[BaseType, str] = {
    BaseType.STR: "string",
    BaseType.INT: "integer",
    BaseType.FLOAT: "number",
    BaseType.BOOL: "boolean",
    BaseType.LIST: "array",
    BaseType.DICT: "object",
    BaseType.TUPLE: "array",
    BaseType.SET: "array",
}

# Reverse mappings for lookup; prioritize first entry in case of duplicates
python_type_to_base_type: Dict[Any, BaseType] = {v: k for k, v in reversed(python_type_mapping.items())}
json_schema_to_base_type: Dict[str, BaseType] = {v: k for k, v in reversed(json_schema_mapping.items())}

def get_python_type(base_type: Union[BaseType, str]) -> Optional[Any]:
    """
    Get the corresponding Python type for a given BaseType.

    Args:
        base_type (BaseType): The BaseType enum value.

    Returns:
        Any: The corresponding Python type.

    Raises:
        KeyError: If no matching Python type is found.

    Examples:
        >>> get_python_type(BaseType.STR)
        <class 'str'>
        >>> get_python_type(BaseType.INT)
        <class 'int'>
        >>> get_python_type(BaseType.FLOAT)
        <class 'float'>
    """
    if isinstance(base_type, str):
        try:
            base_type = BaseType(base_type)
        except ValueError:
            # TODO: ensure types dereferenced previously
            return None
    return python_type_mapping.get(base_type)

def get_json_schema_type(base_type: BaseType) -> str:
    """
    Get the corresponding JSON Schema type for a given BaseType.

    Args:
        base_type (BaseType): The BaseType enum value.

    Returns:
        str: The corresponding JSON Schema type.

    Raises:
        KeyError: If no matching JSON Schema type is found.

    Examples:
        >>> get_json_schema_type(BaseType.STR)
        'string'
        >>> get_json_schema_type(BaseType.INT)
        'integer'
        >>> get_json_schema_type(BaseType.FLOAT)
        'number'
    """
    return json_schema_mapping[base_type]

def from_python_type(py_type: Any) -> BaseType:
    """
    Get the corresponding BaseType for a given Python type.

    Args:
        py_type (Any): The Python type.

    Returns:
        BaseType: The corresponding BaseType enum value.

    Raises:
        ValueError: If no matching BaseType is found.

    Examples:
        >>> from_python_type(str)
        <BaseType.STR: 'str'>
        >>> from_python_type(int)
        <BaseType.INT: 'int'>
        >>> from_python_type(float)
        <BaseType.FLOAT: 'float'>
        >>> from_python_type(complex)
        Traceback (most recent call last):
            ...
        ValueError: No matching BaseType for Python type: <class 'complex'>
    """
    if py_type in python_type_to_base_type:
        return python_type_to_base_type[py_type]
    raise ValueError(f"No matching BaseType for Python type: {py_type}")

def from_json_schema_type(json_type: str) -> BaseType:
    """
    Get the corresponding BaseType for a given JSON Schema type.

    Args:
        json_type (str): The JSON Schema type.

    Returns:
        BaseType: The corresponding BaseType enum value.

    Raises:
        ValueError: If no matching BaseType is found.

    Examples:
        >>> from_json_schema_type("string")
        <BaseType.STR: 'str'>
        >>> from_json_schema_type("integer")
        <BaseType.INT: 'int'>
        >>> from_json_schema_type("number")
        <BaseType.FLOAT: 'float'>
        >>> from_json_schema_type("array")
        <BaseType.LIST: 'list'>
        >>> from_json_schema_type("unknown")
        Traceback (most recent call last):
            ...
        ValueError: No matching BaseType for JSON Schema type: unknown
    """
    if json_type in json_schema_to_base_type:
        return json_schema_to_base_type[json_type]
    raise ValueError(f"No matching BaseType for JSON Schema type: {json_type}")



from abc import ABC, abstractmethod
from collections import abc
from dataclasses import dataclass, field
from pathlib import Path
from typing import Union, Optional, Iterable, Iterator, Any, ClassVar, List, Type, Dict, Tuple, TextIO

from mypy.moduleinspect import ModuleType

from typedlogic import FactMixin, Variable, Implies
from typedlogic.datamodel import Theory, SentenceGroup, PredicateDefinition, Sentence, SentenceGroupType, Term, Exists
from typedlogic.parsers.pyparser.python_parser import PythonParser
from typedlogic.profiles import Profile, UnspecifiedProfile
from typedlogic.pybridge import fact_to_term

ELEMENT = Union[FactMixin, SentenceGroup, Sentence, Theory, PredicateDefinition]

@dataclass
class Solution:
    satisfiable: Optional[bool] = None

@dataclass
class Model:
    """
    A model is a set of ground terms that satisfy a set of axioms.
    """
    description: Optional[str] = None
    source_object: Optional[Any] = None
    ground_terms: List[Term] = field(default_factory=list)

    def iter_retrieve(self, predicate: str, *args) -> Iterator[Term]:
        """
        Retrieve all ground terms with a given predicate.

        :param predicate:
        :return:
        """
        for t in self.ground_terms:
            if t.predicate != predicate:
                continue
            if args:
                is_match = True
                for i in range(len(args)):
                    if args[i] is None:
                        continue
                    if args[i] != t.values[i]:
                        is_match = False
                        break
                if not is_match:
                    continue
            yield t


@dataclass
class Method:
    """
    A method is a way to solve a set of axioms.
    """
    name: str
    description: Optional[str] = None
    is_default: bool = False
    impl_class: Optional[Type] = None

@dataclass
class Solver(ABC):
    """
    A solver is a class that can check a set of axioms for consistency, satisfiability, or some other property.

    This is an abstract class that defines the interface for a solver.

    You can retrieve a specific solver with the `get_solver` function:

        >>> from typedlogic.registry import get_solver
        >>> solver = get_solver("clingo")

    Once you has a solver, you can can add theories, or individual sentences to it:

        >>> from typedlogic.integrations.frameworks.pydantic import FactBaseModel
        >>> class AncestorOf(FactBaseModel):
        ...     ancestor: str
        ...     descendant: str
        >>> solver.add_predicate_definition(PredicateDefinition(predicate="AncestorOf", arguments={'ancestor': str, 'descendant': str}))
        >>> from typedlogic import Term, Variable
        >>> x = Variable("x")
        >>> y = Variable("y")
        >>> z = Variable("z")
        >>> solver.add( (Term("AncestorOf", x, z) & Term("AncestorOf", z, y)) >> Term("AncestorOf", x, y))

    And facts:

        >>> solver.add_fact(AncestorOf(ancestor='p1', descendant='p1a'))
        >>> solver.add_fact(AncestorOf(ancestor='p1a', descendant='p1aa'))
        >>> aa = SentenceGroup(name="transitivity-of-ancestor-of")
        >>> solver.add_sentence_group(aa)

    The `check` method ensures the theory and ground terms (data) are consistent:

        >>> soln = solver.check()
        >>> soln.satisfiable
        True

    You can then query for models:

        >>> model = solver.model()
        >>> for t in model.ground_terms:
        ...     print(t)
        AncestorOf(p1, p1a)
        AncestorOf(p1a, p1aa)
        AncestorOf(p1, p1aa)

    """
    strict: bool = False
    method_name: Optional[str] = None
    methods_supported: ClassVar[Optional[List[Method]]] = None
    profile: ClassVar[Profile] = UnspecifiedProfile()
    assume_closed_world: bool = False

    # TODO: move towards this
    base_theory: Theory = field(default_factory=Theory)

    predicate_definitions: Optional[Dict[str, PredicateDefinition]] = None
    type_definitions: Dict[str, str] = field(default_factory=dict)
    constants: Dict[str, Any] = field(default_factory=dict)
    goals: Optional[List[SentenceGroup]] = None

    @property
    def method(self) -> Method:
        if self.methods_supported is None:
            raise NotImplementedError("Solver must define methods_supported")
        for m in self.methods_supported:
            if self.method_name is None and m.is_default:
                return m
            if m.name == self.method_name:
                return m
        raise ValueError(f"Method {self.method_name} not supported")

    @abstractmethod
    def check(self) -> Solution:
        pass

    def model(self) -> Model:
        return next(self.models())

    @abstractmethod
    def models(self) -> Iterator[Model]:
        pass

    def prove_goals(self, strict=True) -> Iterable[Tuple[Sentence, Optional[bool]]]:
        if not self.check().satisfiable:
            raise ValueError("Cannot prove goals for unsatisfiable theory")
        if not self.goals:
            raise ValueError("No goals to prove")
        for goal_group in self.goals:
            if not goal_group.sentences:
                raise ValueError(f"Goal group {goal_group.name} has no sentences")
            for sentence in goal_group.sentences:
                provable = self.prove(sentence)
                if not provable and strict:
                    raise ValueError(f"Goal {sentence} not provable")
                yield sentence, provable

    def prove_multiple(self, sentences: List[Sentence]) -> Iterable[Tuple[Sentence, Optional[bool]]]:
        if self.check().satisfiable is False:
            raise ValueError("Cannot prove goals for unsatisfiable theory")
        if not sentences:
            raise ValueError("No goals to prove")
        for sentence in sentences:
            provable = self.prove(sentence)
            yield sentence, provable

    def prove(self, sentence: Sentence) -> Optional[bool]:
        """
        Prove a sentence.

        :param sentence:
        :return:
        """
        if isinstance(sentence, Term):
            # Note: the default implementation may be highly ineffecient.
            # it is recommended to override this method in a subclass.
            has_vars = sentence.variables
            cls = type(self)
            new_solver = cls()
            new_solver.add(self.base_theory)
            model = self.model()
            for t in model.iter_retrieve(sentence.predicate):
                if t == sentence:
                    return True
                if has_vars:
                    if t.predicate == sentence.predicate:
                        is_match = True
                        for i in range(len(sentence.values)):
                            arg_val = sentence.values[i]
                            if isinstance(arg_val, Variable):
                                # auto-match (assume existential over whole domain)
                                continue
                            if arg_val != t.values[i]:
                                is_match = False
                                break
                        if is_match:
                            return True
            return False
        if isinstance(sentence, Exists):
            inner = sentence.sentence
            if isinstance(inner, Term):
                return self.prove(inner)
        return None

    def load(self, source: Union[str, Path, TextIO, ModuleType]) -> None:
        """
        Load a theory from a file.

        :param source:
        :return:
        """
        parser = PythonParser()
        if isinstance(source, ModuleType):
            theory = parser.transform(source)
        else:
            theory = parser.parse(source)
        self.add(theory)


    def add(self, element: Union[ELEMENT, Iterable[ELEMENT]]) -> None:
        if isinstance(element, (list, abc.Iterator)):
            for e in element:
                self.add(e)
            return
        if isinstance(element, FactMixin):
            self.add_fact(element)
        elif isinstance(element, SentenceGroup):
            self.add_sentence_group(element)
        elif isinstance(element, Theory):
            self.add_theory(element)
        elif isinstance(element, PredicateDefinition):
            self.add_predicate_definition(element)
        elif isinstance(element, Sentence):
            self.add_sentence(element)
        else:
            raise ValueError(f"Unsupported axiom type: {type(element)}")

    def add_fact(self, fact: FactMixin):
        self.base_theory.ground_terms.append(fact_to_term(fact))

    def add_sentence_group(self, sentence_group: SentenceGroup) -> None:
        self.base_theory.sentence_groups.append(sentence_group)
        if sentence_group.group_type == SentenceGroupType.GOAL:
            if not self.goals:
                self.goals = []
            self.goals.append(sentence_group)
        if sentence_group.sentences:
            for sentence in sentence_group.sentences:
                self.add_sentence(sentence)

    def add_sentence(self, sentence: Sentence) -> None:
        if sentence not in self.base_theory.sentences:
            self.base_theory.sentence_groups.append(SentenceGroup(name="dynamic",
                                                                  sentences=[sentence]))

    def add_predicate_definition(self, predicate_definition: PredicateDefinition) -> None:
        """
        Add a predicate definition to the solver.

        Some solvers do not need predicate definitions (for example, classic prolog systems, as well
        as pure FOL solvers). However, many solvers need some kind of typing information.

        :param predicate_definition:
        :return:
        """
        self.base_theory.predicate_definitions.append(predicate_definition)

    def add_theory(self, theory: Theory) -> None:
        if theory.constants:
            for k, v in theory.constants.items():
                self.constants[k] = v
                self.base_theory.constants[k] = v
        if theory.type_definitions:
            for k, v in theory.type_definitions.items():
                self.type_definitions[k] = v
                self.base_theory.type_definitions[k] = v
        if theory.predicate_definitions:
            for p in theory.predicate_definitions:
                self.add_predicate_definition(p)
        if theory.sentence_groups:
            for aa in theory.sentence_groups:
                self.add_sentence_group(aa)
        if theory.ground_terms:
            for t in theory.ground_terms:
                self.add(t)


    def dump(self) -> str:
        """
        Dump the internal state of the solver as a string.

        :return:
        """
        raise NotImplementedError




from dataclasses import dataclass

from typedlogic import Fact

NodeID = str
Key = str


@dataclass(frozen=True)
class Node(Fact):
    """
    A node is a pointer to a location in a JSON object tree.

    A node has a value, but it is not itself a value.
    """

    loc: NodeID


@dataclass(frozen=True)
class ListNodeHasMember(Fact):
    """True if the node is a list and has the specified member at the given index."""

    loc: NodeID
    offset: int
    member: NodeID

@dataclass(frozen=True)
class ObjectNodeLookup(Fact):
    """True if the node is an object and has the specified key with the given value."""

    loc: NodeID
    key: Key
    member: NodeID

@dataclass(frozen=True)
class NodeStringValue(Fact):
    """The node is a terminal node with a string value."""

    loc: NodeID
    value: str

@dataclass(frozen=True)
class NodeIntValue(Fact):
    """The node is a terminal node with an integer value."""

    loc: NodeID
    value: int

@dataclass(frozen=True)
class NodeFloatValue(Fact):
    """The node is a terminal node with a float value."""

    loc: NodeID
    value: float

@dataclass(frozen=True)
class NodeBooleanValue(Fact):
    """The node is a terminal node with a boolean value."""

    loc: NodeID
    value: bool

@dataclass(frozen=True)
class NodeNullValue(Fact):
    """The node is a terminal node with a null value."""

    loc: NodeID


@dataclass(frozen=True)
class NodeIsList(Fact):
    loc: NodeID

@dataclass(frozen=True)
class NodeIsObject(Fact):
    loc: NodeID

@dataclass(frozen=True)
class NodeIsLiteral(Fact):
    loc: NodeID





import ast
import importlib
import inspect
import types
import typing
from dataclasses import fields
from types import ModuleType
from typing import Any, Dict, List, NewType, Tuple, Type, Union

from typedlogic import Fact, FactMixin, Theory
from typedlogic.datamodel import DefinedType, PredicateDefinition, SentenceGroupType
from typedlogic.parsers.pyparser.python_ast_utils import SentenceGroup, parse_function_def_to_sentence_group
from typedlogic.transformations import ensure_terms_positional, sentences_from_predicate_hierarchy


def translate_module_to_theory(module: ModuleType) -> Theory:
    """
    Translate a module to a Theory object.

        >>> import tests.theorems.mortals as mortals
        >>> theory = translate_module_to_theory(mortals)
        >>> theory.name
        'tests.theorems.mortals'

    :param module:
    :return:
    """
    sgs = get_module_sentence_groups(module)
    pds = get_module_predicate_definitions(module)
    constants, tds = get_module_constants_and_types(module)
    # get the python module name
    theory = Theory(
        name=module.__name__,
        constants=constants,
        type_definitions=tds,
        predicate_definitions=list(pds.values()),
        sentence_groups=sgs,
    )
    ensure_terms_positional(theory)
    new_sentences = sentences_from_predicate_hierarchy(theory)
    for s in new_sentences:
        theory.add(s)
    return theory

# TODO: unused function; decide if we want to allow axiom-level
def get_axioms_ast(cls: Type):
    class_def = ast.parse(ast.unparse(ast.parse(inspect.getsource(cls))))
    for node in ast.walk(class_def):
        if isinstance(node, ast.ClassDef) and node.name == cls.__name__:
            for item in node.body:
                if isinstance(item, ast.FunctionDef) and item.name == 'axioms':
                    return item
    return None


def get_module_sentence_groups(module: Union[ModuleType, str]) -> List[SentenceGroup]:
    """
    Get the AST nodes of all axiom functions in a module.

    :param module: The module to introspect
    :return: A list of AST FunctionDef nodes for axiom functions
    """
    if not isinstance(module, str):
        module = inspect.getsource(module)
    module_ast = ast.parse(module)
    # module_vars = {k: v for k, v in vars(module).items() if not k.startswith('__')}
    sgs = []
    decorator_types = {x.value: x for x in SentenceGroupType}
    for node in module_ast.body:
        if isinstance(node, ast.FunctionDef):
            for decorator in node.decorator_list:
                if isinstance(decorator, ast.Name) and decorator.id in decorator_types:
                    sg = parse_function_def_to_sentence_group(node)
                    sg.group_type = decorator_types[decorator.id]
                    # TODO:
                    #if sg.argument_types is not None:
                    #    sg.argument_types = {k: safe_eval_type(v, module_vars) for k, v in sg.argument_types.items()}
                    sgs.append(sg)
                    break
    return sgs


def get_module_predicate_classes(module:  ModuleType) -> Dict[str, Type]:
    """
    Get all classes defined in a module.

    :param module: The module to introspect
    :return: A dictionary of class names to class types
    """
    module_vars = {k: v for k, v in vars(module).items() if not k.startswith('__')}
    classes = {}
    for name, obj in module_vars.items():
        if not inspect.isclass(obj):
            continue
        if inspect.isabstract(obj):
            continue
        if obj in {Fact, FactMixin}:
            continue
        if not issubclass(obj, FactMixin):
            continue
        classes[name] = obj
    return classes

JSON_SCHEMA_TYPE_MAP = {
    "string": "str",
    "number": "float",
    "integer": "int",
}

def introspect_attributes(cls: Type) -> dict[str, Any]:
    # https://stackoverflow.com/questions/69090253/how-to-iterate-over-attributes-of-dataclass-in-python
    try:

        import pydantic
        if issubclass(cls, pydantic.BaseModel):
            schema = cls.model_json_schema()
            r = {}
            for p, p_schema in schema['properties'].items():
                def _parse_type(s) -> DefinedType:
                    if 'anyOf' in s:
                        return [_parse_type(x) for x in s['anyOf']]
                    return s.get('type', 'str')
                t = _parse_type(p_schema)
                if isinstance(t, str):
                    r[p] = JSON_SCHEMA_TYPE_MAP.get(t, t)
                else:
                    # TODO
                    r[p] = JSON_SCHEMA_TYPE_MAP['string']
            return r
    except ImportError:
        pass    # Get all attributes of the class
    if fields(cls):
        r = {field.name: field.type.__name__ for field in fields(cls)}
        return r

    attributes = inspect.getmembers(cls)

    # Filter out class variables and built-in attributes
    non_classvar_attributes = [
        attr for attr in attributes
        if not isinstance(attr[1], type)
           and not attr[0].startswith('__')
           and not attr[0].startswith('_')
           and not inspect.ismethod(attr[1])
           and not inspect.isfunction(attr[1])
    ]

    return {attr[0]: attr[1] for attr in non_classvar_attributes}

def get_module_constants(module: ModuleType) -> Dict[str, Any]:
    """
    Get all constants defined in a module.

    :param module: The module to introspect
    :return: A dictionary of constant names to constant values
    """
    module_vars = {k: v for k, v in vars(module).items() if not k.startswith('__')}
    constants = {}
    for name, obj in module_vars.items():
        if not inspect.isclass(obj) and not inspect.isfunction(obj):
            constants[name] = obj
    return constants


def is_union(t):
    if str(t).startswith('typing.Union'):
        return True
    # For Python 3.10+
    if hasattr(types, "UnionType"):
        return isinstance(t, types.UnionType)

    # For earlier versions
    return typing.get_origin(t) is Union

def get_module_constants_and_types(module: ModuleType) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    module_vars = {k: v for k, v in vars(module).items() if not k.startswith('__')}
    constants = {}
    types: Dict[str, DefinedType] = {}
    for name, obj in module_vars.items():
        if not inspect.isfunction(obj):
            if inspect.isclass(obj):
                t = obj.__name__
                if t != name:
                    types[name] = t
            elif is_union(obj):
                t = obj.__name__
                if t != name:
                    types[name] = [x.__name__ for x in typing.get_args(obj)]
            elif isinstance(obj, NewType):
                st = obj.__supertype__
                if isinstance(st, type):
                    types[name] = st.__name__
                else:
                    types[name] = str(st)
            else:
                constants[name] = obj
    return constants, types

def get_module_predicate_definitions(module: ModuleType) -> Dict[str, PredicateDefinition]:
    """
    Get all predicate classes defined in a module.

    Assume a module with the following (Pydantic) class:

        >>> from pydantic import BaseModel
        >>> from typedlogic import FactMixin
        >>> class PersonWithAge(BaseModel, FactMixin):
        ...   name: str
        ...   age: int

    This is in one of the tests:

        >>> import tests.theorems.types_example as types_example
        >>> pds = get_module_predicate_definitions(types_example)
        >>> list(pds.keys())
        ['PersonWithAge', 'Adult', 'StageAge']
        >>> pd = pds['PersonWithAge']
        >>> pd.predicate
        'PersonWithAge'
        >>> pd.arguments
        {'name': 'str', 'age': 'int'}

    :param module:
    :return:
    """
    cls_name_map = get_module_predicate_classes(module)
    pds = {}
    parent_map = {}
    cls_to_pd = {}
    for name, cls in cls_name_map.items():
        pd = PredicateDefinition(name, introspect_attributes(cls))
        pds[name] = pd
        # get python parent classes
        parent_map[pd.predicate] = cls.__bases__
        cls_to_pd[cls] = pd
    for predicate, pd in pds.items():
        pd.parents = []
        for parent in parent_map[predicate]:
            if parent in cls_to_pd:
                pd.parents.append(cls_to_pd[parent].predicate)

    return pds


def safe_eval_type(type_str: str, globals_dict: Dict[str, Any]) -> Any:
    """
    Safely evaluate a type string, considering potential imports.

    :param type_str: The string representation of the type to evaluate
    :type type_str: str
    :param globals_dict: A dictionary of global variables, including imports
    :type globals_dict: Dict[str, Any]
    :return: The evaluated type
    :rtype: Any

    :raises ValueError: If the type cannot be evaluated safely
    """
    try:
        # First, try to evaluate as is (handles built-in types and already imported types)
        return eval(type_str, globals_dict)
    except NameError:
        # If NameError, it might be due to an import. Let's try to import it.
        module_parts = type_str.split('.')
        for i in range(1, len(module_parts)):
            try:
                module_name = '.'.join(module_parts[:i])
                globals_dict[module_parts[0]] = importlib.import_module(module_name)
                return eval(type_str, globals_dict)
            except ImportError:
                continue
        raise ValueError(f"Cannot safely evaluate type: {type_str}")



import logging
from dataclasses import dataclass, field
from typing import Any, ClassVar, Iterator, Optional

import clingo
from clingo import Control, SymbolType

from typedlogic.datamodel import NotInProfileError, Sentence, Term
from typedlogic.profiles import (
    AllowsComparisonTerms,
    AnswerSetProgramming,
    MixedProfile,
    MultipleModelSemantics,
    Profile,
)
from typedlogic.solver import Model, Solution, Solver
from typedlogic.transformations import PrologConfig, as_prolog, to_horn_rules

logger = logging.getLogger(__name__)

@dataclass
class ClingoSolver(Solver):
    """
    A solver that uses clingo.

    [clingo](https://potassco.org/clingo/) is an ASP system to ground and solve logic programs, it is
    part of the Potsdam Answer Set Solving Collection (Potassco; [potassco.org/](https://potassco.org/)).


        >>> from typedlogic.integrations.frameworks.pydantic import FactBaseModel
        >>> class AncestorOf(FactBaseModel):
        ...     ancestor: str
        ...     descendant: str
         >>> from typedlogic import SentenceGroup, PredicateDefinition
        >>> solver = ClingoSolver()
        >>> solver.add_predicate_definition(PredicateDefinition(predicate="AncestorOf", arguments={'ancestor': str, 'descendant': str}))
        >>> solver.add_fact(AncestorOf(ancestor='p1', descendant='p1a'))
        >>> solver.add_fact(AncestorOf(ancestor='p1a', descendant='p1aa'))

        >>> aa = SentenceGroup(name="transitivity-of-ancestor-of")
        >>> solver.add_sentence_group(aa)
        >>> soln = solver.check()

    This solver does not implement the open-world assumption.

        >>> from typedlogic.profiles import OpenWorld
        >>> solver.profile.impl(OpenWorld)
        False

    """

    exec_name: str = field(default="clingo")
    profile: ClassVar[Profile] = MixedProfile(AnswerSetProgramming(), AllowsComparisonTerms(), MultipleModelSemantics())
    ctl : Optional[Control] = None


    def _clauses(self) -> Iterator[str]:
        negation_symbol = "not" if self.assume_closed_world else "-"
        prolog_config = PrologConfig(disjunctive_datalog=True, double_quote_strings=True, negation_symbol=negation_symbol, allow_nesting=False)
        for sentence in self.base_theory.sentences + self.base_theory.ground_terms:
            if not isinstance(sentence, Sentence):
                raise ValueError(f"Expected Sentence, got {sentence}")
            rules = []
            try:
                for rule in to_horn_rules(sentence, allow_disjunctions_in_head=True, allow_goal_clauses=True):
                    rules.append(rule)
                    # yield as_prolog(rule, config=prolog_config)
            except NotInProfileError as e:
                logger.info(f"Skipping sentence {sentence} due to {e}")
            for rule in rules:
                try:
                    yield as_prolog(rule, config=prolog_config)
                except NotInProfileError as e:
                    logger.info(f"Skipping sentence {sentence} due to {e}")



    def models(self) -> Iterator[Model]:
        ctl = Control(["0"])
        predicate_name_map = {pd.predicate.lower(): pd.predicate for pd in self.base_theory.predicate_definitions}
        for clause in self._clauses():
            ctl.add(clause)

        # Ground the program
        ctl.ground([("base", [])])
        # Solve the program
        with ctl.solve(yield_=True) as handle:
            for clingo_model in handle:
                facts = []
                for atom in clingo_model.symbols(shown=True):
                    p = predicate_name_map.get(atom.name, atom.name)
                    def _v(sym: clingo.Symbol) -> Any:
                        if sym.type == SymbolType.String:
                            return str(sym.string)
                        if sym.type == SymbolType.Number:
                            return sym.number
                        return sym
                    term = Term(p, *[_v(a) for a in atom.arguments])
                    if not atom.positive:
                        term = ~term
                    facts.append(term)

                model = Model(ground_terms=facts)
                yield model

    def check(self) -> Solution:
        models = list(self.models())
        sat = len(models) > 0
        return Solution(satisfiable=sat)

    def dump(self) -> str:
        s = ""
        for clause in self._clauses():
            s += f"{clause}\n"
        return s




"""
Command-line interface definitions using typer.
"""
from pathlib import Path
from typing import List, Optional

import typer

from typedlogic.parsers.pyparser.python_parser import PythonParser
from typedlogic.registry import get_compiler, get_solver

app = typer.Typer()



@app.command()
def convert(
        input_files: List[Path] = typer.Argument(..., exists=True, dir_okay=False, readable=True),
        output_format: str = typer.Option(None, "--output-format", "-t", help="Output format"),
        output_file: Optional[Path] = typer.Option(None, "--output-file", "-o", help="Output file path")
):
    """
    Convert from one logic form to another.
    """
    parser = PythonParser()
    theory = parser.parse(input_files[0])
    if len(input_files) > 1:
        for input_file in input_files[1:]:
            facts_theory = parser.parse(input_file)
            for s in facts_theory.sentences:
                theory.add(s)

    compiler = get_compiler(output_format or "z3sexpr")
    result = compiler.compile(theory)

    if output_file:
        with open(output_file, 'w') as f:
            f.write(result)
        typer.echo(f"Conversion result written to {output_file}")
    else:
        typer.echo(result)


@app.command()
def solve(
        input_file: Path = typer.Argument(..., exists=True, dir_okay=False, readable=True),
        solver: str = typer.Option(None, help="Solver to use"),
        output_file: Optional[Path] = typer.Option(None, "--output-file", "-o", help="Output file path for the solution")
):
    """
    Solve using the specified solver.
    """
    parser = PythonParser()
    theory = parser.parse(input_file)
    solver_instance = get_solver(solver or "souffle")

    solver_instance.add(theory)
    solution = solver_instance.check()

    result = f"Satisfiable: {solution.satisfiable}\n"
    if solution.satisfiable:
        model = solver_instance.model()
        result += "Model:\n"
        for fact in model.ground_terms:
            result += f"{fact}\n"

    if output_file:
        with open(output_file, 'w') as f:
            f.write(result)
        typer.echo(f"Solution written to {output_file}")
    else:
        typer.echo(result)


if __name__ == "__main__":
    app()



from abc import ABC
from typing import Tuple, Mapping, Any, Dict, Type

import pydantic
from pydantic import BaseModel

from typedlogic import Sentence, Term
from typedlogic.datamodel import Extension


class FactMixin(Extension, ABC):
    """
    Mixin class for facts

    You can use this if you want to effectively inherit from Fact, but want to base your
    classes using your own base class, or something like the Pydantic BaseModel.
    """
    def to_model_object(self) -> Sentence:
        return fact_to_term(self)


class Fact(FactMixin, ABC):
    """
    Abstract base class for grounded facts (unit clauses).

    This class should be subclassed to make domain-specific Fact subclasses.

    E.g

        >>> from pydantic import BaseModel
        >>> class Person(BaseModel, FactMixin):
        ...     name: str
        ...     age: int

    You can then instantiate individual facts:

        >>> p = Person(name="Alice", age=30)

    You can introspect the fact to get the predicate and arguments:

        >>> fact_predicate(p)
        'Person'

        >>> assert fact_arg_py_types(p) == {'name': str, 'age': int}

    """
    pass


def fact_args(fact: FactMixin) -> Tuple[str, ...]:
    """Return the arguments of a predicate"""
    return tuple(vars(fact).keys())


def fact_arg_map(fact: FactMixin) -> Mapping[str, Any]:
    """Return the arguments of a predicate"""
    return vars(fact)


def fact_arg_py_types(fact: FactMixin) -> Dict[str, Type]:
    """
    Introspect the predicate class to get typing information
    """
    return {k: type(v) for k, v in vars(fact).items()}


def fact_arg_values(fact: FactMixin) -> Tuple[Any, ...]:
    """Return the predicate of a sentence"""
    return tuple(vars(fact).values())


def fact_predicate(fact: FactMixin) -> str:
    """Return the predicate of a sentence"""
    return type(fact).__name__


def fact_to_term(fact: FactMixin) -> Term:
    """Convert a fact to a term"""
    return Term(fact_predicate(fact), fact_arg_map(fact))



from typing import List

from typedlogic import PredicateDefinition, Term
from typedlogic.datamodels.typesystem import get_python_type


def make_terms(rows: List[List[str]], pd: PredicateDefinition) -> List[Term]:
    """
    Make terms from a list of rows
    """
    terms = []
    for i, pd_arg in enumerate(pd.arguments.values()):
        py_type = get_python_type(pd_arg)
        if not py_type:
            continue
        if py_type == str:
            continue
        for j, row in enumerate(rows):
            rows[j][i] = py_type(row[i])
    pred = pd.predicate
    for row in rows:
        if not pd.arguments:
            terms.append(Term(pred))
        else:
            # TODO: coerce ints
            terms.append(Term(pred, *row))
    return terms



from dataclasses import dataclass

from typedlogic import Fact

ElementID = str
ClassDefinitionID = ElementID
SlotDefinitionID = ElementID

@dataclass(frozen=True)
class Element(Fact):
    id: ElementID

@dataclass(frozen=True)
class ClassDefinition(Element):
    pass

@dataclass(frozen=True)
class TypeDefinition(Element):
    pass

@dataclass(frozen=True)
class EnumDefinition(Element):
    pass

@dataclass(frozen=True)
class SlotDefinition(Element):
    pass

@dataclass(frozen=True)
class SchemaDefinition(Element):
    pass

@dataclass(frozen=True)
class TreeRoot(Fact):
    cls: ElementID

@dataclass(frozen=True)
class ClassSlot(Fact):
    """
    Maps to class.slots
    """

    cls: ElementID
    slot: ElementID

@dataclass(frozen=True)
class Identifier(Fact):
    cls: ElementID
    slot_expression: ElementID

@dataclass(frozen=True)
class Multivalued(Fact):
    slot_expression: ElementID

@dataclass(frozen=True)
class Required(Fact):
    slot_expression: ElementID

@dataclass(frozen=True)
class IsA(Fact):
    element: ElementID
    parent: ElementID

@dataclass(frozen=True)
class Mixin(Fact):
    element: ElementID
    parent: ElementID

@dataclass(frozen=True)
class Attribute(Fact):
    cls: ElementID
    slot_expression: ElementID

@dataclass(frozen=True)
class SlotUsage(Fact):
    # instead make specific?
    cls: ElementID
    slot_expression: ElementID



from typing import ClassVar, Optional, Union

from typedlogic import Theory
from typedlogic.compiler import Compiler, ModelSyntax
from typedlogic.integrations.solvers.z3.z3_solver import Z3Solver


class Z3Compiler(Compiler):
    """
    Implementation of the Compiler interface for Z3
    """

    def compile(self, theory: Theory, syntax: Optional[Union[str, ModelSyntax]] = None,  **kwargs) -> str:
        solver = Z3Solver()
        solver.add(theory)
        if syntax is None:
            syntax = ModelSyntax.SEXPR
        if syntax == ModelSyntax.SEXPR:
            return solver.wrapped_solver.sexpr()
        elif syntax == ModelSyntax.FUNCTIONAL:
            return str(solver.wrapped_solver)
        else:
            raise ValueError(f"Unknown syntax: {syntax}")

class Z3SExprCompiler(Z3Compiler):
    """
    Implementation of the Compiler interface for Z3 with S-Expression syntax
    """

    default_suffix: ClassVar[str] = "z3.sxpr"

    def compile(self, theory: Theory, syntax: Optional[Union[str, ModelSyntax]] = None, **kwargs) -> str:
        return super().compile(theory, ModelSyntax.SEXPR, **kwargs)

class Z3FunctionalCompiler(Z3Compiler):
    """
    Implementation of the Compiler interface for Z3 with functional syntax
    """

    default_suffix: ClassVar[str] = "z3.fun"

    def compile(self, theory: Theory,syntax: Optional[Union[str, ModelSyntax]] = None,  **kwargs) -> str:
        return super().compile(theory, ModelSyntax.FUNCTIONAL, **kwargs)



from dataclasses import dataclass, field
from typing import Dict, List


@dataclass
class Node:
    id: int
    edges: List[int] = field(default_factory=list)
    index: int = -1
    lowlink: int = -1
    on_stack: bool = False

@dataclass
class Graph:
    nodes: Dict[int, Node] = field(default_factory=dict)
    stack: List[Node] = field(default_factory=list)
    index: int = 0
    sccs: List[List[int]] = field(default_factory=list)

    def add_edge(self, u: int, v: int) -> None:
        if u not in self.nodes:
            self.nodes[u] = Node(u)
        if v not in self.nodes:
            self.nodes[v] = Node(v)
        self.nodes[u].edges.append(v)

    def tarjan(self) -> List[List[int]]:
        for node in self.nodes.values():
            if node.index == -1:
                self._strong_connect(node)
        return self.sccs

    def _strong_connect(self, v: Node) -> None:
        v.index = self.index
        v.lowlink = self.index
        self.index += 1
        self.stack.append(v)
        v.on_stack = True

        for w_id in v.edges:
            w = self.nodes[w_id]
            if w.index == -1:
                self._strong_connect(w)
                v.lowlink = min(v.lowlink, w.lowlink)
            elif w.on_stack:
                v.lowlink = min(v.lowlink, w.index)

        if v.lowlink == v.index:
            scc = []
            while True:
                w = self.stack.pop()
                w.on_stack = False
                scc.append(w.id)
                if w == v:
                    break
            self.sccs.append(scc)

def main():
    g = Graph()
    edges = [(0, 1), (1, 2), (2, 0), (1, 3), (3, 4), (4, 3)]
    for u, v in edges:
        g.add_edge(u, v)

    sccs = g.tarjan()
    print("Strongly Connected Components:")
    for i, scc in enumerate(sccs, 1):
        print(f"SCC {i}: {scc}")

if __name__ == "__main__":
    main()



from dataclasses import dataclass
from typing import Iterator

import rdflib
from rdflib.term import Node

from typedlogic import FactMixin, axiom


@dataclass
class Triple(FactMixin):
    subject: Node
    predicate: Node
    object: Node


@dataclass
class NodeIntegerValue(FactMixin):
    node: Node
    value: int


@dataclass
class NodeStringValue(FactMixin):
    node: Node
    value: str


@dataclass
class IsIRI(FactMixin):
    node: Node


@dataclass
class IsBlank(FactMixin):
    node: Node


@axiom
def _sp_iris(s: Node, p: Node, o: Node):
    # assert Triple("A", "B", "C")
    if Triple(s, p, o):
        IsIRI(p)
    if Triple(s, p, o):
        IsIRI(s) | IsBlank(s)


def generate_sentences(g: rdflib.Graph) -> Iterator[Triple]:
    # yield Triple(str(OWL.Thing), str(RDF.type), str(OWL.Thing))
    for s, p, o in g:
        if isinstance(o, rdflib.URIRef):
            # yield Triple(str(s), str(p), str(o))
            yield Triple(s, p, o)
        else:
            # TODO
            # yield Triple(str(s), str(p), str(o))
            yield Triple(s, p, o)



import logging
from dataclasses import dataclass, field
from typing import Any, ClassVar, Iterable, Iterator, List, Optional, Tuple

import llm
import yaml

from typedlogic import Sentence
from typedlogic.profiles import MixedProfile, OpenWorld, Profile, Unrestricted
from typedlogic.registry import get_compiler
from typedlogic.solver import Model, Solution, Solver

logger = logging.getLogger(__name__)

SYSTEM = """
Your task is to determine whether logical statements follow from a given set of axioms.
I will provide you with an initial list of axioms. I will then give a numbered set of
sentences. You will return whether that sentence follows from the axioms.

Give results as a simple yaml object like this:

```yaml
provable: [1, 3, 5]
not_provable: [2, 4, 6]
```

Do not provide proofs, just the yaml object.
"""

TEMPLATE = """
Background theory:

{program}

Which of the following are entailed:

{goals}

Answer:
"""

@dataclass
class LLMSolver(Solver):
    """
    A solver that uses an LLM.

    This is for research purposes only. LLM output can be reliable, slow, and expensive,
    and should not be used as a replacement for a deterministic logic-based solver.

    """

    model_name: str = field(default="gpt-4o")
    fol_syntax: str = field(default="fol")
    profile: ClassVar[Profile] = MixedProfile(Unrestricted(), OpenWorld())

    def models(self) -> Iterator[Model]:
        r = self.check()
        if r.satisfiable:
            yield Model()

    def prove(self, sentence: Sentence) -> Optional[bool]:
        results = list(self.prove_multiple([sentence]))
        return results[0][1]

    def prove_multiple(self, sentences: List[Sentence]) -> Iterable[Tuple[Sentence, Optional[bool]]]:
        compiler = get_compiler(self.fol_syntax)
        program = compiler.compile(self.base_theory)
        model = llm.get_model(self.model_name)
        enumerated_goals = dict(enumerate(sentences, 1))
        enumerated_goals_compiled = {i: compiler.compile_sentence(s) for i, s in enumerated_goals.items()}
        goals = "\n".join([f"{i}: {s}" for i, s in enumerated_goals_compiled.items()])
        prompt = TEMPLATE.format(program=program, goals=goals)
        #print(f"SYSTEM={SYSTEM}")
        #print(f"PROMPT={prompt}")
        response = model.prompt(prompt, system=SYSTEM)
        print(f"RESPONSE={response.text()}")
        obj = self.parse_response(response.text())
        for i in obj["provable"]:
            yield enumerated_goals[int(i)], True
        for i in obj["not_provable"]:
            yield enumerated_goals[int(i)], False

    def parse_response(self, text: str) -> Any:
        if "```" in text:
            text = text.split("```")[1].strip()
            if text.startswith("yaml"):
                text = text[5:].strip()
        return yaml.safe_load(text)


    def check(self) -> Solution:
        return Solution(satisfiable=None)



"""
Mappings for builtin functions and datatypes.
"""
import operator as op
from typing import Mapping, Callable

NUMERIC_BUILTINS: Mapping[str, Callable] = {
    'ge': op.ge,
    'gt': op.gt,
    'le': op.le,
    'lt': op.lt,
    'eq': op.eq,
    'ne': op.ne,
    'add': op.add,
    'sub': op.sub,
    'mul': op.mul,
    'truediv': op.truediv,
    'pow': op.pow,
    'xor': op.xor,
    'neg': op.neg,
}

NAME_TO_INFIX_OP: Mapping[str, str] = {
'add': '+',
    'sub': '-',
    'mul': '*',
    'truediv': '/',
    'floordiv': '//',
    'mod': '%',
    'pow': '**',
    'lshift': '<<',
    'rshift': '>>',
    'or': '|',
    'xor': '^',
    'and': '&',
    'matmul': '@',
    # Comparison operators
    'eq': '==',
    'ne': '!=',
    'lt': '<',
    'le': '<=',
    'gt': '>',
    'ge': '>=',
    'is': 'is',
    'is_not': 'is not',
    'in': 'in',
    'not_in': 'not in',
}


"""
OWL-DL Top Level Classes and Axioms.

In this integration, OWL-DL is treated as a subset of FOL, consisting of (a) Unary predicate
definitions (classes) and (b) Binary predicate definitions (Object and Data Properties).

This module also provides Pythonic integration through top-level classes that can be inherited
to defined OWL ontologies through Python classes. These Python classes can be treated as normal python
classes, they can be instantiated to create an ABox. They can also be translated into FOL axioms,
where they can be used for reasoning over ABoxes using a Solver.

This module is pure python and does not attempt to parse or serialize OWL - for that,
an integration with a framework such as Py-horned-owl is recommended.
"""
from abc import ABC
from dataclasses import dataclass, field
from decimal import Decimal
from functools import wraps
from typing import Any, Callable, ClassVar, Iterator, List, Mapping, Optional, Tuple, Type, Union

from typedlogic import And, Exists, Fact, Forall, Iff, Implies, Not, Or, Sentence, Term, Variable

I = Variable("I")
J = Variable("J")
K = Variable("K")
P = Variable("P")
C = Variable("C")
D = Variable("D")



def as_list(e) -> List:
    if e == None:
        return []
    if isinstance(e, list):
        return e
    return [e]

def _conjunction(sentences: List[Sentence]) -> Sentence:
    if len(sentences) == 0:
        raise ValueError("Cannot create conjunction of zero sentences")
    if len(sentences) == 1:
        return sentences[0]
    return And(*sentences)

def instance_of(inst_var: Variable, ce: "ClassExpression") -> Sentence:
    if isinstance(ce, AnonymousClassExpression):
        return ce.as_fol() or And()
    if isinstance(ce, str):
        return Term(ce, inst_var)
    return Term(ce.__name__, inst_var)


def instance_of_op(inst_var1: Variable, inst_var2: Variable, ce: "ObjectPropertyExpression") -> Optional[Sentence]:
    if isinstance(ce, InverseObjectProperty):
        return instance_of_op(inst_var2, inst_var1, ce.property)
    return Term(ce.__name__, inst_var1, inst_var2)


def instance_of_dp(inst_var1: Variable, inst_var2: Variable, ce: "DataPropertyExpression") -> Optional[Sentence]:
    return Term(ce.__name__, inst_var1, inst_var2)

IRI = str




@dataclass(frozen=True)
class Thing(Fact):
    """
    The top level class in OWL-DL. Everything is a Thing.

    For your object model to be included in the OWL-DL mapping, all your classes
    representing individuals should inherit from Thing.

    Basic Example:

    ```python
    class Person(Thing):
        '''A person'''

    class Pet(Thing):
        '''A pet'''
    ```

    Formally, OWL Classes are unary predicates, and instances are unary ground terms:

    ```python
    facts = [Person("P1"), Pet("Fido")]
    ```

    You can create more complex defined classes, using OWL Frame-style class definitions:

    ```python
    class HasPet(TopObjectProperty):
        '''A property that relates a person to a pet'''
        domain = Person
        range = Pet

    class PetOwner(Person):
        '''A person that owns a pet'''
        equivalent_to = Person & ObjectSomeValuesFrom(HasPet, Pet)
    ```

    A thing class can have ClassVars `equivalent_to`, `subclass_of`, ...
    Note these are class-level.

    We can create and reason over instances and facts:

    ```python
    facts = [Person("P1"), Pet("Fido"), HasPet("P1", "Fido")]
    reasoner = OWLReasoner()
    reasoner.init_from_file(Path(__file__))
    for f in facts:
        reasoner.add(f)
    model = reasoner.model()
    for t in model.ground_terms:
        print(t)
    ```

    Results will include:

    ```
    PetOwner(P1)
    ```

    Formally, the Thing class is a subclass of a typed-logic Fact, which instantiates an
    OWLCLass python metaclass:

    ```mermaid
    classDiagram
    class Thing {
        IRI iri
    }
    Fact <|-- Thing
    Thing --> `OWLCLass[[metaclass]]` : instance_of
    `OWLCLass[[metaclass]]` --> "*" ClassExpression : subclass_of
    `OWLCLass[[metaclass]]` --> "*" ClassExpression : equivalent_to
    `OWLCLass[[metaclass]]` --> "*" ClassExpression : disjoint_with
    `OWLCLass[[metaclass]]` --> "*" ClassExpression : disjoint_union_of
    ```

    """

    iri: IRI

    subclass_of: ClassVar[Optional["ClassExpression"]] = None
    equivalent_to: ClassVar[Optional["ClassExpression"]] = None
    disjoint_with: ClassVar[Optional["ClassExpression"]] = None
    disjoint_union_of: ClassVar[Optional[List["ClassExpression"]]] = None

    @classmethod
    def axioms_iter(cls) -> Iterator["Axiom"]:
        for parent in cls.__bases__:
            if issubclass(parent, Thing):
                yield SubClassOf(cls, parent)
        yield from [SubClassOf(cls, e) for e in as_list(cls.subclass_of)]
        yield from [DisjointClasses(cls, e) for e in as_list(cls.disjoint_with)]
        yield from [EquivalentClasses(cls, e) for e in as_list(cls.equivalent_to)]
        if cls.disjoint_union_of:
            yield DisjointUnion(cls, *cls.disjoint_union_of)

    @classmethod
    def axioms(cls) -> List["Axiom"]:
        return list(cls.axioms_iter())

    @classmethod
    def to_sentences(cls) -> List[Sentence]:
        return [a for a in [axiom.as_fol() for axiom in cls.axioms()] if a is not None]

@dataclass(frozen=True)
class TopObjectProperty(Fact):
    """
    The top object property.

    Example:
    -------
    ```python
    class Person(Thing):
        '''A person'''

    class HasParent(TopObjectProperty):
        '''A property that relates a person to their parent'''
        domain = Person
        range = Person

    class HasAncestor(HasParent):
        '''A property that relates a person to their ancestor'''
        transitive = True
    ```

    Formally, OWL Object Properties are binary predicates, and instances are binary ground terms:

    ```python
    facts = [HasParent("P1", "P2"), HasAncestor("P1", "P3")]
    ```

    Formally, the TopObjectProperty class is a subclass of a 2-ary typed-logic Fact, which instantiates an
    OWLObjectProperty python metaclass:

    ```mermaid
    classDiagram
    class TopObjectProperty {
        IRI subject
        IRI object
    }
    Fact <|-- TopObjectProperty
    TopObjectProperty --> `OWLObjectProperty[[metaclass]]` : instance_of
    `OWLObjectProperty[[metaclass]]` --> "*" ObjectPropertyExpression : subproperty_of
    `OWLObjectProperty[[metaclass]]` --> "*" PropertyExpressionChain : subproperty_chain
    `OWLObjectProperty[[metaclass]]` --> "*" ClassExpression : domain
    `OWLObjectProperty[[metaclass]]` --> "*" ClassExpression : range

    class `OWLObjectProperty[[metaclass]]` {
       Boolean transitive
       Boolean symmetric
       Boolean asymmetric
       Boolean reflexive
      Boolean irreflexive
      Boolean functional
      Boolean inverse_functional
    }
    ```



    """

    subject: IRI
    object: IRI

    transitive: ClassVar[Optional[bool]] = None
    symmetric: ClassVar[Optional[bool]] = None
    asymmetric: ClassVar[Optional[bool]] = None
    reflexive: ClassVar[Optional[bool]] = None
    irreflexive: ClassVar[Optional[bool]] = None
    functional: ClassVar[Optional[bool]] = None
    inverse_functional: ClassVar[Optional[bool]] = None
    subproperty_of: ClassVar[Optional["ObjectPropertyExpression"]] = None
    subproperty_chain: ClassVar[Optional["PropertyExpressionChain"]] = None
    equivalent_to: ClassVar[Optional["ObjectPropertyExpression"]] = None
    disjoint_with: ClassVar[Optional["ObjectPropertyExpression"]] = None
    inverse_of: ClassVar[Optional["ObjectPropertyExpression"]] = None
    domain: ClassVar[Optional["ClassExpression"]] = None
    range: ClassVar[Optional["ClassExpression"]] = None

    @classmethod
    def some(cls, ce: "ClassExpression") -> "ObjectSomeValuesFrom":
        return ObjectSomeValuesFrom(cls, ce)

    @classmethod
    def only(cls, ce: "ClassExpression") -> "ObjectAllValuesFrom":
        return ObjectAllValuesFrom(cls, ce)

    @classmethod
    def value(cls, i: "Individual") -> "ObjectHasValue":
        return ObjectHasValue(cls, i)

    @classmethod
    def axioms_iter(cls) -> Iterator["Axiom"]:
        for parent in cls.__bases__:
            if issubclass(parent, TopObjectProperty):
                yield SubObjectPropertyOf(cls, parent)
        yield from [SubObjectPropertyOf(cls, e) for e in as_list(cls.subproperty_of)]
        yield from [SubObjectPropertyOf(chain, cls) for chain in as_list(cls.subproperty_chain)]
        yield from [ObjectPropertyDomain(cls, e) for e in as_list(cls.domain)]
        yield from [ObjectPropertyRange(cls, e) for e in as_list(cls.range)]
        if cls.transitive:
            yield from [TransitiveObjectProperty(cls)]
        if cls.symmetric:
            yield from [SymmetricObjectProperty(cls)]
        if cls.asymmetric:
            yield from [AsymmetricObjectProperty(cls)]
        if cls.reflexive:
            yield from [ReflexiveObjectProperty(cls)]
        if cls.irreflexive:
            yield from [IrreflexiveObjectProperty(cls)]
        if cls.functional:
            yield from [FunctionalObjectProperty(cls)]
        if cls.inverse_functional:
            yield from [InverseFunctionalObjectProperty(cls)]
        yield from [EquivalentObjectProperties(cls, e) for e in as_list(cls.equivalent_to)]
        yield from [DisjointObjectProperties(cls, e) for e in as_list(cls.disjoint_with)]
        if cls.inverse_of:
            yield from [InverseObjectProperties(cls, cls.inverse_of)]

    @classmethod
    def axioms(cls) -> List["Axiom"]:
        return list(cls.axioms_iter())

    @classmethod
    def to_sentences(cls) -> List[Sentence]:
        return [a for a in [axiom.as_fol() for axiom in cls.axioms()] if a is not None]


ObjectProperty = Type[TopObjectProperty]


@dataclass(frozen=True)
class TopDataProperty(Fact):
    """
    The top data property.

    Example:
    -------
    ```python
    class Person(Thing):
        '''A person'''

    class HasAge(TopDataProperty):
        '''A property that relates a person to their age'''
        domain = Person
        range = int
    ```

    Formally, OWL Data Properties are binary predicates, and instances are binary ground terms:

    ```python
    facts = [HasAge("P1", 25)]
    ```

    Formally, the TopDataProperty class is a subclass of a 2-ary typed-logic Fact, which instantiates an
    OWLDataProperty python metaclass:

    ```mermaid
    classDiagram
    class TopDataProperty {
        IRI subject
        Union[str, int, float, bool, Decimal] object
    }
    Fact <|-- TopDataProperty
    TopDataProperty --> `OWLDataProperty[[metaclass]]` : instance_of
    `OWLDataProperty[[metaclass]]` --> "*" DataPropertyExpression : subproperty_of
    `OWLDataProperty[[metaclass]]` --> "*" ClassExpression : domain
    `OWLDataProperty[[metaclass]]` --> "*" DataRange : range
    class `OWLDataProperty[[metaclass]]` {
         Boolean functional
     }
     ```

    """

    subject: IRI
    object: Union[str, int, float, bool, Decimal]

    functional: ClassVar[Optional[bool]] = None
    subproperty_of: ClassVar[Optional["DataPropertyExpression"]] = None
    equivalent_to: ClassVar[Optional["DataPropertyExpression"]] = None
    disjoint_with: ClassVar[Optional["DataPropertyExpression"]] = None
    domain: ClassVar[Optional["ClassExpression"]] = None
    range: ClassVar[Optional["DataRange"]] = None

    @classmethod
    def axioms_iter(cls) -> Iterator["Axiom"]:
        for parent in cls.__bases__:
            if issubclass(parent, TopDataProperty):
                yield SubDataPropertyOf(cls, parent)
        yield from [SubDataPropertyOf(cls, e) for e in as_list(cls.subproperty_of)]
        yield from [DataPropertyDomain(cls, e) for e in as_list(cls.domain)]
        yield from [DataPropertyRange(cls, e) for e in as_list(cls.range)]
        if cls.functional:
            yield from [FunctionalDataProperty(cls)]
        yield from [EquivalentDataProperties(cls, e) for e in as_list(cls.equivalent_to)]
        yield from [DisjointDataProperties(cls, e) for e in as_list(cls.disjoint_with)]

    @classmethod
    def axioms(cls) -> List["Axiom"]:
        return list(cls.axioms_iter())

    @classmethod
    def to_sentences(cls) -> List[Sentence]:
        return [a for a in [axiom.as_fol() for axiom in cls.axioms()] if a is not None]

@dataclass
class Literal(ABC):
    """
    A literal.
    """

    pass

@dataclass
class SimpleLiteral:
    """
    A simple literal.
    """

    literal: str

@dataclass
class LanguageLiteral:
    """
    A language literal.
    """

    literal: str
    lang: str

@dataclass
class DatatypeLiteral:
    """
    A datatype literal.
    """

    literal: str
    datatype_iri: IRI

# classes and properties are declared as python classes (types)



# support str for forward references
Class = Union[Type[Thing], str]





DataProperty = Type[TopDataProperty]

Datatype = Union[IRI, Type[str], Type[int], Type[float], Type[bool]]
Individual = IRI
ClassExpression = Union[Class, "AnonymousClassExpression"]
ObjectPropertyExpression = Union[ObjectProperty, "InverseObjectProperty"]
DataPropertyExpression = DataProperty
# DataRange = Union[DataIntersectionOf, DataUnionOf, DataComplementOf, DataOneOf, DatatypeRestriction, Datatype]
DataRange = Union[Datatype, "AnonymousDataRange"]

@dataclass
class Axiom(ABC):
    """
    Abstract base class for axioms.

    Axioms can be declared implicitly, in "Frame-style" class definitions, or in
    an `__axioms__` module variable
    """

    annotations: ClassVar[Optional[Mapping[str, Any]]] = None

    #@abstractmethod
    def as_fol(self) -> Optional[Sentence]:
        return None

@dataclass
class SubClassOf(Axiom):
    """
    A sub-class-of axiom.

    Axioms can be declared explicitly:

    ```python
    class Person(Thing):
        '''A person'''

    class Employee(Thing):
        '''A person that is employed'''

    __axioms__ = [ SubClassOf(Employee, Person) ]
    ```

    However, it is more idiomatic to declare axioms implicitly in Frame-style class definitions:

    ```python
    class Person(Thing):
        '''A person'''

    class Employee(Thing):
        '''A person that is employed'''
        subclass_of = Person
    ```

    Note that when the superclass is a named class, the more pythonic equivalent is preferred:

    ```python
    class Employee(Person):
        '''A person that is employed'''
    ```

    Using an explicit class var is necessary when the superclass is an expression:

    ```python
    class Organization(Thing):
        '''An organization'''

    class EmployedBy(TopObjectProperty):
        domain = Employee
        range = Organization

    class Employee(Person):
        '''A person that is employed'''
        subclass_of = ObjectSomeValuesFrom(EmployedBy, Organization)
    ```

    UML:

    ```mermaid
    classDiagram
    AnonymousClassExpression <|-- SubClassOf
    SubClassOf --> "*" ClassExpression : sub
    SubClassOf --> "*" ClassExpression : sup
    ```

    """

    sub: ClassExpression
    sup: ClassExpression

    def as_fol(self) -> Optional[Sentence]:
        return Forall([I], Implies(instance_of(I, self.sub), instance_of(I, self.sup)))


@dataclass
class PropertyExpressionChain:
    """
    A chain of property expressions.

    Example:
    -------
    ```python
    class HasParent(TopObjectProperty):
        '''A property that relates a person to their parent'''
        domain = Person
        range = Person

    class HasGrandparent(TopObjectProperty):
        '''A property that relates a person to their grandparent'''
        subproperty_chain = PropertyExpressionChain(HasParent, HasParent)
    ```

    UML:

    ```
    classDiagram
    PropertyExpressionChain --> "*" ObjectPropertyExpression : chain
    ```

    """

    chain: Tuple[ObjectPropertyExpression, ...]

    def __init__(self, *chain: ObjectPropertyExpression):
        self.chain = chain

@dataclass
class SubObjectPropertyOf(Axiom):
    """
    A sub-object-property-of axiom.

    Example of explicit declaration:

    ```python
    class HasParent(TopObjectProperty):
        '''A property that relates a person to their parent'''
        domain = Person
        range = Person

    class HasGrandparent(TopObjectProperty):
        '''A property that relates a person to their grandparent'''

    __axioms__ = [ SubObjectPropertyOf(HasGrandparent, HasParent) ]
    ```

    Normally, the more Frame-style equivalent is preferred:

    ```python
    class HasParent(TopObjectProperty):
        '''A property that relates a person to their parent'''
        domain = Person
        range = Person

    class HasGrandparent(TopObjectProperty):
        '''A property that relates a person to their grandparent'''
        subproperty_of = HasParent
    ```

    When the parent is a named class, the more pythonic equivalent is preferred:

    ```python
    class HasGrandparent(HasParent):
        '''A property that relates a person to their grandparent'''
    ```

    UML:

    ```mermaid
    classDiagram
    AnonymousClassExpression <|-- SubObjectPropertyOf
    SubObjectPropertyOf --> "*" ObjectPropertyExpressionOrPropertyExpressionChain : sub
    SubObjectPropertyOf --> "*" ObjectPropertyExpression : sup
    ```

    """

    sub: Union[ObjectPropertyExpression, PropertyExpressionChain]
    sup: ObjectPropertyExpression

    def as_fol(self) -> Optional[Sentence]:
        if isinstance(self.sup, PropertyExpressionChain):
            raise ValueError(f"Cannot have a chain as a super property - did you accidentally invert? {self}")
        if isinstance(self.sub, PropertyExpressionChain):
            chain_len = len(self.sub.chain)
            inst_vars = [Variable(f"J{i}") for i in range(chain_len + 1)]
            conjs = []
            for i in range(chain_len):
                conjs.append(instance_of_op(inst_vars[i], inst_vars[i + 1], self.sub.chain[i]))
            return Forall(inst_vars, Implies(And(*conjs), instance_of_op(inst_vars[0], inst_vars[-1], self.sup)))
        else:
            return Forall([P, I, J], Implies(instance_of_op(I, J, self.sub), instance_of_op(I, J, self.sup)))

@dataclass
class TransitiveObjectProperty(Axiom):
    """
    A transitive-object-property axiom.

    Example of explicit declaration:

    ```python
    class HasAncestor(TopObjectProperty):
        pass

    __axioms__ = [ TransitiveObjectProperty(HasAncestor) ]
    ```

    Normally, the more Frame-style equivalent is preferred:

    ```python
    class HasAncestor(TopObjectProperty):
        transitive = True
    ```

    UML:

    ```mermaid
    classDiagram
    Axiom <|-- TransitiveObjectProperty
    TransitiveObjectProperty --> "1" ObjectPropertyExpression : first
    ```

    """

    first: ObjectPropertyExpression

    def as_fol(self) -> Optional[Sentence]:
        return Forall([I, J, K], Implies(And(instance_of_op(I, J, self.first), instance_of_op(J, K, self.first)), instance_of_op(I, K, self.first)))

@dataclass
class InverseObjectProperties(Axiom):
    """
    An inverse-object-properties axiom.

    Example of explicit declaration:

    ```python
    class HasParent(TopObjectProperty):
        '''A property that relates a person to their parent'''
        domain = Person
        range = Person

    class HasChild(TopObjectProperty):
        '''A property that relates a person to their child'''
        domain = Person
        range = Person

    __axioms__ = [ InverseObjectProperties(HasParent, HasChild) ]
    ```

    Normally, the more Frame-style equivalent is preferred:

    ```python
    class HasParent(TopObjectProperty):
        '''A property that relates a person to their parent'''
        domain = Person
        range = Person
        inverse_of = HasChild
    ```

    UML:

    ```mermaid
    classDiagram
    Axiom <|-- InverseObjectProperties
    InverseObjectProperties --> "1" ObjectPropertyExpression : first
    InverseObjectProperties --> "1" ObjectPropertyExpression : second
    ```
    """

    first: ObjectPropertyExpression
    second: ObjectPropertyExpression

    def as_fol(self) -> Optional[Sentence]:
        return Forall([I, J], Iff(instance_of_op(I, J, self.first), instance_of_op(J, I, self.second)))

@dataclass
class SymmetricObjectProperty(Axiom):
    """
    A symmetric-object-property axiom.

    Example of explicit declaration:

    ```python
    class HasSibling(TopObjectProperty):
        '''A property that relates a person to their sibling'''
        domain = Person
        range = Person

    __axioms__ = [ SymmetricObjectProperty(HasSibling) ]
    ```

    Normally, the more Frame-style equivalent is preferred:

    ```python
    class HasSibling(TopObjectProperty):
        '''A property that relates a person to their sibling'''
        domain = Person
        range = Person
        symmetric = True
    ```

    UML:

    ```mermaid
    classDiagram
    Axiom <|-- SymmetricObjectProperty
    SymmetricObjectProperty --> "1" ObjectPropertyExpression : first
    ```

    """

    first: ObjectPropertyExpression

    def as_fol(self) -> Optional[Sentence]:
        return Forall([I, J], Iff(instance_of_op(I, J, self.first), instance_of_op(J, I, self.first)))

@dataclass
class AsymmetricObjectProperty(Axiom):
    """
    An asymmetric-object-property axiom.

    Example of explicit declaration:

    ```python
    class HasParent(TopObjectProperty):
        '''A property that relates a person to their parent'''
        domain = Person
        range = Person

    __axioms__ = [ AsymmetricObjectProperty(HasParent) ]
    ```

    Normally, the more Frame-style equivalent is preferred:

    ```python
    class HasParent(TopObjectProperty):
        '''A property that relates a person to their parent'''
        domain = Person
        range = Person
        asymmetric = True
    ```

    UML:

    ```mermaid
    classDiagram
    Axiom <|-- AsymmetricObjectProperty
    AsymmetricObjectProperty --> "1" ObjectPropertyExpression : first
    ```

    """

    first: ObjectPropertyExpression

    def as_fol(self) -> Optional[Sentence]:
        return Forall([I, J], Implies(instance_of_op(I, J, self.first), Not(instance_of_op(J, I, self.first))))

@dataclass
class ReflexiveObjectProperty(Axiom):
    """
    A reflexive-object-property axiom.

    Example of explicit declaration:

    ```python
    class RelatedTo(TopObjectProperty):
        '''A property that relates a thing to itself'''
        domain = Thing
        range = Thing

    __axioms__ = [ ReflexiveObjectProperty(RelatedTo) ]
    ```

    Normally, the more Frame-style equivalent is preferred:

    ```python
    class RelatedTo(TopObjectProperty):
        '''A property that relates a thing to itself'''
        domain = Thing
        range = Thing
        reflexive = True
    ```

    In practice, relfexive axioms are too strong, but we include them for completeness.

    UML:

    ```mermaid
    classDiagram
    Axiom <|-- ReflexiveObjectProperty
    ReflexiveObjectProperty --> "1" ObjectPropertyExpression : first
    ```

    """

    first: ObjectPropertyExpression

    def as_fol(self) -> Optional[Sentence]:
        return Forall([I, C], Implies(instance_of(I, Thing), instance_of_op(I, I, self.first)))

@dataclass
class IrreflexiveObjectProperty(Axiom):
    """
    An irreflexive-object-property axiom.

    Example of explicit declaration:

    ```python

    class HasParent(TopObjectProperty):
        '''A property that relates a person to their parent'''
        domain = Person
        range = Person

    __axioms__ = [ IrreflexiveObjectProperty(HasParent) ]
    ```

    Normally, the more Frame-style equivalent is preferred:

    ```python
    class HasParent(TopObjectProperty):
        '''A property that relates a person to their parent'''
        domain = Person
        range = Person
        irreflexive = True
    ```

    """

    first: ObjectPropertyExpression

    def as_fol(self) -> Optional[Sentence]:
        refl_s = instance_of_op(I, I, self.first)
        if refl_s:
            return Not(Exists([I], refl_s))
        return None

@dataclass
class FunctionalObjectProperty(Axiom):
    """
    A functional-object-property axiom.

    Example of explicit declaration:

    ```python

    class HasBiologicalMother(TopObjectProperty):
        '''A property that relates a person to their biological mother'''
        domain = Person
        range = Person

    __axioms__ = [ FunctionalObjectProperty(HasBiologicalMother) ]
    ```

    Normally, the more Frame-style equivalent is preferred:

    ```python
    class HasBiologicalMother(TopObjectProperty):
        '''A property that relates a person to their biological mother'''
        domain = Person
        range = Person
        functional = True
    ```

    UML:

    ```mermaid
    classDiagram
    Axiom <|-- FunctionalObjectProperty
    FunctionalObjectProperty --> "1" ObjectPropertyExpression : first
    ```

    """

    first: ObjectPropertyExpression

    def as_fol(self) -> Optional[Sentence]:
        return Forall([I, J, K], Implies(And(instance_of_op(I, J, self.first), instance_of_op(I, K, self.first)), Term("eq", J, K)))

@dataclass
class InverseFunctionalObjectProperty(Axiom):
    """
    An inverse-functional-object-property axiom.

    Example of explicit declaration:

    ```python
    class BiologicalMotherOf(TopObjectProperty):
        '''A property that relates a person to their biological mother'''
        domain = Person
        range = Person

    __axioms__ = [ InverseFunctionalObjectProperty(BiologicalMotherOf) ]

    ```

    Normally, the more Frame-style equivalent is preferred:

    ```python
    class BiologicalMotherOf(TopObjectProperty):
        '''A property that relates a person to their biological mother'''
        domain = Person
        range = Person
        inverse_functional = True
    ```

    UML:

    ```mermaid
    classDiagram
    Axiom <|-- InverseFunctionalObjectProperty
    InverseFunctionalObjectProperty --> "1" ObjectPropertyExpression : first
    ```
    """

    first: ObjectPropertyExpression

    def as_fol(self) -> Optional[Sentence]:
        return Forall([I, J, K], Implies(And(instance_of_op(I, J, self.first), instance_of_op(K, J, self.first)), Term("eq", I, K)))

@dataclass
class FunctionalDataProperty(Axiom):
    """
    A functional-data-property axiom.

    Example of explicit declaration:

    ```python
    class HasAge(TopDataProperty):
        '''A property that relates a person to their age'''
        domain = Person
        range = int

    __axioms__ = [ FunctionalDataProperty(HasAge) ]
    ```

    Normally, the more Frame-style equivalent is preferred:

    ```python
    class HasAge(TopDataProperty):
        '''A property that relates a person to their age'''
        domain = Person
        range = int
        functional = True
    ```

    UML:

    ```mermaid
    classDiagram
    Axiom <|-- FunctionalDataProperty
    FunctionalDataProperty --> "1" DataPropertyExpression : first
    ```

    """

    first: DataPropertyExpression

    def as_fol(self) -> Optional[Sentence]:
        return Forall([I, J, K], Implies(And(instance_of_dp(I, J, self.first), instance_of_dp(I, K, self.first)), Term("eq", J, K)))

@dataclass
class SubDataPropertyOf(Axiom):
    """
    A sub-data-property-of axiom.

    Example of explicit declaration:

    ```python
    class HasAge(TopDataProperty):
        '''A property that relates a thing to their age'''
        range = int

    class PersonAge(TopDataProperty):
        '''A property that relates a person to their age'''
        domain = Person

    __axioms__ = [ SubDataPropertyOf(PersonAge, HasAge) ]
    ```

    Normally, the more Frame-style equivalent is preferred:

    ```python
    class PersonAge(TopDataProperty):
        '''A property that relates a person to their age'''
        domain = Person
        range = int
        subproperty_of = HasAge
    ```

    When the parent is a named class, the more pythonic equivalent is preferred:

    ```python
    class PersonAge(HasAge):
        '''A property that relates a person to their age'''
    ```

    UML:

    ```mermaid
    classDiagram
    Axiom <|-- SubDataPropertyOf
    SubDataPropertyOf --> "*" DataPropertyExpression : sub
    SubDataPropertyOf --> "*" DataPropertyExpression : sup
    ```
    """

    sub: DataPropertyExpression
    sup: DataPropertyExpression

    def as_fol(self) -> Optional[Sentence]:
        return Forall([P, I, J], Implies(instance_of_dp(I, J, self.sub), instance_of_dp(I, J, self.sup)))



@dataclass
class EquivalentClasses(Axiom):
    """
    An equivalent-classes axiom.

    Example of explicit declaration:

    ```python
    class Food(Thing):
        pass

    class Beverage(Thing):
        pass

    class MenuItem(Thing):
        pass

    __axioms__ = [ EquivalentClasses(MenuItem, ObjectUnionOf(Food, Beverage)) ]
    ```

    Normally, the more Frame-style equivalent is preferred:

    ```python
    class MenuItem(Thing):
        equivalent_to = ObjectUnionOf(Food, Beverage)
    ```

    UML:

    ```mermaid
    classDiagram
    Axiom <|-- EquivalentClasses
    EquivalentClasses --> "*" ClassExpression : operands
    ```
    """

    operands: Tuple[ClassExpression, ...] = field(default_factory=tuple)

    def __init__(self, *operands: ClassExpression):
        self.operands = operands

    def as_fol(self) -> Optional[Sentence]:
        num_ops = len(self.operands)
        sentences: List[Sentence] = []
        for i in range(num_ops - 1):
            for j in range(i + 1, num_ops):
                sentences.append(Iff(instance_of(I, self.operands[i]), instance_of(I, self.operands[j])))
        return Forall([I], _conjunction(sentences))

@dataclass
class EquivalentObjectProperties(Axiom):
    """
    An equivalent-object-properties axiom.

    Example of explicit declaration:

    ```python
    class HasParent(TopObjectProperty):
        '''A property that relates a person to their parent'''
        domain = Person
        range = Person

    class HasChild(TopObjectProperty):
        '''A property that relates a person to their child'''
        domain = Person
        range = Person

    __axioms__ = [ EquivalentObjectProperties(HasParent, ObjectInverseOf(HasChild)) ]
    ```

    Normally, the more Frame-style equivalent is preferred:

    ```python
    class HasChild(TopObjectProperty):
        equivalent_to = ObjectInverseOf(HasParent)
    ```

    In this particular case, a simpler more conventional way to declare the inverse is:

    ```python
    class HasChild(TopObjectProperty):
        range = Person
        domain = Person
        inverse_of = HasParent
    ```

    UML:

    ```mermaid
    classDiagram
    Axiom <|-- EquivalentObjectProperties
    EquivalentObjectProperties --> "*" ObjectPropertyExpression : operands
    ```

    """

    operands: Tuple[ObjectPropertyExpression, ...] = field(default_factory=tuple)

    def __init__(self, *operands: ObjectPropertyExpression):
        self.operands = operands

    def as_fol(self) -> Optional[Sentence]:
        num_ops = len(self.operands)
        sentences: List[Sentence] = []
        for i in range(num_ops - 1):
            for j in range(i + 1, num_ops):
                sentences.append(Iff(instance_of_op(I, J, self.operands[i]), instance_of_op(I, J, self.operands[j])))
        return Forall([I, J], _conjunction(sentences))

@dataclass
class EquivalentDataProperties(Axiom):
    """
    An equivalent-data-properties axiom.

    UML:

    ```mermaid
    classDiagram
    Axiom <|-- EquivalentDataProperties
    EquivalentDataProperties --> "*" DataPropertyExpression : operands
    ```
    """

    operands: Tuple[DataPropertyExpression, ...] = field(default_factory=tuple)

    def __init__(self, *operands: DataPropertyExpression):
        self.operands = operands

    def as_fol(self) -> Optional[Sentence]:
        num_ops = len(self.operands)
        sentences: List[Sentence] = []
        for i in range(num_ops - 1):
            for j in range(i + 1, num_ops):
                sentences.append(Iff(instance_of_dp(I, J, self.operands[i]), instance_of_dp(I, J, self.operands[j])))
        return Forall([I, J], _conjunction(sentences))

@dataclass
class DisjointObjectProperties(Axiom):
    """
    A disjoint-object-properties axiom.

    UML:

    ```mermaid
    classDiagram
    Axiom <|-- DisjointObjectProperties
    DisjointObjectProperties --> "*" ObjectPropertyExpression : operands
    ```
    """

    operands: Tuple[ObjectPropertyExpression, ...] = field(default_factory=tuple)

    def __init__(self, *operands: ObjectPropertyExpression):
        self.operands = operands

    def as_fol(self) -> Optional[Sentence]:
        num_ops = len(self.operands)
        sentences: List[Sentence] = []
        for i in range(num_ops - 1):
            for j in range(i + 1, num_ops):
                sentences.append(Not(And(instance_of_op(I, J, self.operands[i]), instance_of_op(I, J, self.operands[j]))))
        return Not(Exists([I, J], Or(*sentences)))

@dataclass
class DisjointDataProperties(Axiom):
    """
    A disjoint-data-properties axiom.

    UML:

    ```mermaid
    classDiagram
    Axiom <|-- DisjointDataProperties
    DisjointDataProperties --> "*" DataPropertyExpression : operands
    ```
    """

    operands: Tuple[DataPropertyExpression, ...] = field(default_factory=tuple)

    def __init__(self, *operands: DataPropertyExpression):
        self.operands = operands

    def as_fol(self) -> Optional[Sentence]:
        num_ops = len(self.operands)
        sentences: List[Sentence] = []
        for i in range(num_ops - 1):
            for j in range(i + 1, num_ops):
                sentences.append(Not(And(instance_of_dp(I, J, self.operands[i]), instance_of_dp(I, J, self.operands[j]))))
        return Not(Exists([I, J], Or(*sentences)))

@dataclass
class DisjointClasses(Axiom):
    """
    A disjoint-classes axiom.

    Example of explicit declaration:

    ```python
    class Food(Thing):
        pass

    class Beverage(Thing):
        pass

    __axioms__ = [ DisjointClasses(Food, Beverage) ]
    ```

    Sometimes, the more Frame-style equivalent is preferred:

    ```python
    class Food(Thing):
        pass

    class Beverage(Thing):
        disjoint_with = Food
    ```

    However, the choice of which element of the tuple is the first class is arbitrary,
    so declaration of a top level axiom may be simpler (especially when more than
    two classes are members, or class expressions are involved).

    UML:

    ```mermaid
    classDiagram
    Axiom <|-- DisjointClasses
    DisjointClasses --> "*" ClassExpression : operands
    ```
    """

    operands: Tuple[ClassExpression, ...] = field(default_factory=tuple)

    def __init__(self, *operands: ClassExpression):
        self.operands = operands

    def as_fol(self) -> Optional[Sentence]:
        num_ops = len(self.operands)
        sentences: List[Sentence] = []
        for i in range(num_ops - 1):
            for j in range(i + 1, num_ops):
                sentences.append(Not(And(instance_of(I, self.operands[i]), instance_of(I, self.operands[j]))))
        return Not(Exists([I], Or(*sentences)))

@dataclass
class DisjointUnion(Axiom):
    """
    A disjoint-union axiom.

    Example of explicit declaration:

    ```python
    class Food(Thing):
        pass

    class Beverage(Thing):
        pass

    class MenuItem(Thing):
        pass

    __axioms__ = [ DisjointUnion(MenuItem, Food, Beverage) ]
    ```

    Sometimes, the more Frame-style equivalent is preferred:

    ```python
    class Food(Thing):
        pass

    class Beverage(Thing):
        disjoint_union_of = MenuItem
    ```

    UML:

    ```mermaid
    classDiagram
    Axiom <|-- DisjointUnion
    DisjointUnion --> "1" ClassExpression : first
    DisjointUnion --> "*" ClassExpression : operands
    ```
    """

    first: Class
    operands: Tuple[ClassExpression, ...] = field(default_factory=tuple)

    def __init__(self, first: Class, *operands: ClassExpression):
        self.first = first
        self.operands = operands

    def as_fol(self) -> Optional[Sentence]:
        eq = Forall([I], Iff(instance_of(I, self.first), Or(*[instance_of(I, op) for op in self.operands])))
        num_ops = len(self.operands)
        disj: List[Sentence] = []
        for i in range(num_ops - 1):
            for j in range(i + 1, num_ops):
                disj.append(Not(And(instance_of(I, self.operands[i]), instance_of(I, self.operands[j]))))
        return And(eq, Not(Exists([I], Or(*disj))))

@dataclass
class AnonymousClassExpression(ABC):
    """
    An base class for anonymous class expressions.
    """

    def as_fol(self) -> Optional[Sentence]:
        raise NotImplementedError(f"{self} Must implement as_fol method")

@dataclass
class InverseObjectProperty:
    """
    An inverse object property expression.

    Example:
    -------
    ```python
    class HasParent(TopObjectProperty):
        '''A property that relates a person to their parent'''
        domain = Person
        range = Person

    class HasChild(TopObjectProperty):
        '''A property that relates a person to their child'''

    __axioms__ = [ EquivalentObjectProperties(HasParent, InverseObjectProperty(HasChild)) ]
    ```

    UML:

    ```mermaid
    classDiagram
    InverseObjectProperty --> "1" ObjectPropertyExpression : property
    ```

    """

    property: ObjectProperty

    def as_fol(self) -> Optional[Sentence]:
        raise AssertionError("Not to be called directly")

@dataclass
class ObjectIntersectionOf(AnonymousClassExpression):
    """
    A ClassExpression that is the intersection of a list of ClassExpressions.

    Example of explicit axiom declaration:

    ```python
    class Pizza(Thing):
        pass

    class Cheesy(Thing):
        pass

    class CheesyPizza(Thing):
        pass

    __axioms__ = [ EquivalentClasses(CheesyPizza, ObjectIntersectionOf(Pizza, Cheesy)) ]
    ```

    It is more idiomatic to declare axioms implicitly in Frame-style class definitions:

    ```python
    class CheesyPizza(Thing):
        equivalent_to = ObjectIntersectionOf(Pizza, Cheesy)
    ```

    UML:

    ```mermaid
    classDiagram
    AnonymousClassExpression <|-- ObjectIntersectionOf
    ObjectIntersectionOf --> "*" ClassExpression : operands
    ```

    """

    operands: Tuple[ClassExpression, ...] = field(default_factory=tuple)

    def __init__(self, *operands: ClassExpression):
        self.operands = operands

    def as_fol(self) -> Sentence:
        return And(*[instance_of(I, op) for op in self.operands])

@dataclass
class ObjectUnionOf(AnonymousClassExpression):
    """
    A ClassExpression that is the union of a list of ClassExpressions.

    Example of explicit axiom declaration:

    ```python
    class Pizza(Thing):
        pass

    class CheeseOnToast(Thing):
        pass

    class Pizzaesque(Thing)
        pass

    __axioms__ = [ EquivalentClasses(Pizzaesque, ObjectUnionOf(Pizza, CheeseOnToast)) ]
    ```

    It is more idiomatic to declare axioms implicitly in Frame-style class definitions

    ```python
    class Pizzaesque(Thing):
        equivalent_to = ObjectUnionOf(Pizza, CheeseOnToast)
    ```

    UML:

    ```mermaid
    classDiagram
    AnonymousClassExpression <|-- ObjectUnionOf
    ObjectUnionOf --> "*" ClassExpression : operands
    ```
    """

    operands: Tuple[ClassExpression, ...] = field(default_factory=tuple)

    def __init__(self, *operands: ClassExpression):
        self.operands = operands

    def as_fol(self) -> Optional[Sentence]:
        return Or(*[instance_of(I, op) for op in self.operands])

@dataclass
class ObjectComplementOf(AnonymousClassExpression):
    """
    A ClassExpression that is the complement of a ClassExpression.

    Example of explicit axiom declaration:

    ```python
    class Pizza(Thing):
        pass

    class MeatPizza(Thing):
        pass

    class VegetarianPizza(Thing):
        pass

    __axioms__ = [ EquivalentClasses(VegetarianPizza, ObjectIntersectionOf(Pizza, ObjectComplementOf(MeatPizza))) ]
    ```

    It is more idiomatic to declare axioms implicitly in Frame-style class definitions

    ```python
    class VegetarianPizza(Thing):
        equivalent_to = ObjectIntersectionOf(Pizza, ObjectComplementOf(MeatPizza))
    ```

    UML:

    ```mermaid
    classDiagram
    AnonymousClassExpression <|-- ObjectComplementOf
    ObjectComplementOf --> "1" ClassExpression : first
    ```
    """

    first: ClassExpression

    def as_fol(self) -> Optional[Sentence]:
        return Not(instance_of(I, self.first))

@dataclass
class ObjectOneOf(AnonymousClassExpression):
    """
    A ClassExpression that is the set of individuals.

    Example of explicit axiom declaration:

    ```python
    class Unit(Thing):
        pass

    inch = Unit("inch")
    cm = Unit("cm")
    m = Unit("m")

    class LengthUnit(Unit):
        pass

    __axioms__ = [ EquivalentClasses(LengthUnit, ObjectOneOf(inch, cm, m)) ]
    ```

    It is more idiomatic to declare axioms implicitly in Frame-style class definitions

    ```python
    class LengthUnit(Unit):
        equivalent_to = ObjectOneOf(inch, cm, m)
    ```

    UML:

    ```mermaid
    classDiagram
    AnonymousClassExpression <|-- ObjectOneOf
    ObjectOneOf --> "*" Individual : operands
    ```
    """

    operands: Tuple[Individual, ...] = field(default_factory=tuple)

    def as_fol(self) -> Optional[Sentence]:
        return Or(*[Term("eq", I, ind) for ind in self.operands])

@dataclass
class ObjectSomeValuesFrom(AnonymousClassExpression):
    """
    A ClassExpression representing an existential restriction.

    Example of explicit axiom declaration:

    ```python
    class HasChil(TopObjectProperty):
        '''A property that relates a person to their child'''
        domain = Person
        range = Person

    class Parent(Thing):
        pass

    __axioms__ = [ EquivalentClasses(Parent, ObjectSomeValuesFrom(HasChild, Person)) ]
    ```

    It is more idiomatic to declare axioms implicitly in Frame-style class definitions:

    ```python
    class Parent(Thing):
        equivalent_to = ObjectSomeValuesFrom(HasChild, Person)
    ```

    UML:

    ```mermaid
    classDiagram
    AnonymousClassExpression <|-- ObjectSomeValuesFrom
    ObjectSomeValuesFrom --> "1" ObjectPropertyExpression : ope
    ObjectSomeValuesFrom --> "1" ClassExpression : bce
    ```

    """

    ope: ObjectPropertyExpression
    bce: ClassExpression

    def as_fol(self) -> Optional[Sentence]:
        return Exists([J], And(instance_of_op(I, J, self.ope), instance_of(J, self.bce)))

@dataclass
class ObjectAllValuesFrom(AnonymousClassExpression):
    """
    A ClassExpression representing a universal restriction.

    Example of explicit axiom declaration:

    ```python
    class Pizza(Thing):
        pass

    class Eats(TopObjectProperty):
        domain = Person
        range = Thing

    class Pizzavore(TopObjectProperty):
        pass

    __axioms__ = [ EquivalentClasses(Pizzavore, ObjectAllValuesFrom(Eats, Pizza)) ]
    ```

    It is more idiomatic to declare axioms implicitly in Frame-style class definitions

    ```python
    class Pizzavore(Thing):
        equivalent_to = ObjectAllValuesFrom(Eats, Pizza)
    ```

    UML:

    ```mermaid
    classDiagram
    AnonymousClassExpression <|-- ObjectAllValuesFrom
    ObjectAllValuesFrom --> "1" ObjectPropertyExpression : ope
    ObjectAllValuesFrom --> "1" ClassExpression : bce
    ```

    """

    ope: ObjectPropertyExpression
    bce: ClassExpression

    def as_fol(self) -> Optional[Sentence]:
        return Forall([J], Implies(instance_of_op(I, J, self.ope), instance_of(J, self.bce)))

@dataclass
class ObjectHasValue(AnonymousClassExpression):
    """
    A ClassExpression representing a value restriction.

    Example of explicit axiom declaration:

    ```python

    class Topping(Thing):
        pass

    class Pizza(Thing):
        pass

    class ToppedWith(TopObjectProperty):
        domain = Pizza
        range = Topping

    class MushroomPizza(Topping):
        pass

    __axioms__ = [ EquivalentClasses(MushroomPizza, ObjectHasValue(ToppedWith, Topping("mushroom")) ]
    ```

    It is more idiomatic to declare axioms implicitly in Frame-style class definitions:

    ```python
    class MushroomPizza(Topping):
        equivalent_to = ObjectHasValue(ToppedWith, Topping("mushroom"))
    ```

    """

    ope: ObjectPropertyExpression
    i: Individual

    def as_fol(self) -> Optional[Sentence]:
        return Exists([J], And(instance_of_op(I, J, self.ope), Term("eq", J, self.i)))

@dataclass
class ObjectHasSelf(AnonymousClassExpression):
    """
    A ClassExpression representing a self restriction.

    Example of explicit axiom declaration:

    ```python
    class Person(Thing):
        pass

    class Loves(TopObjectProperty):
        domain = Person
        range = Thing

    class Narcissist(Thing):
        pass

    __axioms__ = [ EquivalentClasses(Narcissist, ObjectHasSelf(Loves)) ]
    ```

    It is more idiomatic to declare axioms implicitly in Frame-style class definitions:

    ```python
    class Narcissist(Thing):
        equivalent_to = ObjectHasSelf(Loves)
    ```

    UML:

    ```mermaid
    classDiagram
    AnonymousClassExpression <|-- ObjectHasSelf
    ObjectHasSelf --> "1" ObjectPropertyExpression :

    """

    ope: ObjectPropertyExpression

    def as_fol(self) -> Optional[Sentence]:
        return instance_of_op(I, I, self.ope)

@dataclass
class ObjectMinCardinality(AnonymousClassExpression):
    """
    A ClassExpression representing a minimum cardinality restriction.

    Example of explicit axiom declaration:

    ```python
    class HasChild(TopObjectProperty):
        '''A property that relates a person to their child'''
        domain = Person
        range = Person

    class ParentOfTwoPlus(Thing):
        pass

    __axioms__ = [ EquivalentClasses(ParentOfTwoPlus, ObjectMinCardinality(2, HasChild, Person)) ]
    ```

    It is more idiomatic to declare axioms implicitly in Frame-style class definitions:

    ```python
    class ParentOfTwoPlus(Thing):
        equivalent_to = ObjectMinCardinality(2, HasChild, Person)
    ```

    UML:

    ```mermaid
    classDiagram
    AnonymousClassExpression <|-- ObjectMinCardinality
    ObjectMinCardinality --> "1" int : n
    ObjectMinCardinality --> "1" ObjectPropertyExpression : ope
    ObjectMinCardinality --> "1" ClassExpression : bce
    ```

    """

    n: int
    ope: ObjectPropertyExpression
    bce: ClassExpression

    # TODO: Aggregates
    #def as_fol(self) -> Optional[Sentence]:
    #   return Forall([J], Implies(instance_of_op(I, J, self.ope), instance_of(J, self.bce)))

@dataclass
class ObjectMaxCardinality(AnonymousClassExpression):
    """
    A ClassExpression representing a maximum cardinality restriction.

    Example of explicit axiom declaration:

    ```python
    class HasChild(TopObjectProperty):
        '''A property that relates a person to their child'''
        domain = Person
        range = Person

    class ParentOfOne(Thing):
        pass

    __axioms__ = [ EquivalentClasses(ParentOfOne, ObjectMinCardinality(0, HasChild, Person) & ObjectMaxCardinality(1, HasChild, Person)) ]
    ```

    It is more idiomatic to declare axioms implicitly in Frame-style class definitions:

    ```python
    class ParentOfOne(Thing):
        equivalent_to = ObjectMinCardinality(0, HasChild, Person) & ObjectMaxCardinality(1, HasChild, Person)
    ```

    UML:

    ```mermaid
    classDiagram
    AnonymousClassExpression <|-- ObjectMaxCardinality
    ObjectMaxCardinality --> "1" int : n
    ObjectMaxCardinality --> "1" ObjectPropertyExpression : ope
    ObjectMaxCardinality --> "1" ClassExpression : bce
    ```

    """

    n: int
    ope: ObjectPropertyExpression
    bce: ClassExpression

    # TODO: Aggregates

@dataclass
class ObjectExactCardinality(AnonymousClassExpression):
    """
    A ClassExpression representing an exact cardinality restriction.

    Example of explicit axiom declaration:

    ```python
    class HasChild(TopObjectProperty):
        '''A property that relates a person to their child'''
        domain = Person
        range = Person

    class ParentOfOne(Thing):
        pass

    __axioms__ = [ EquivalentClasses(ParentOfOne, ObjectExactCardinality(1, HasChild, Person)) ]
    ```

    It is more idiomatic to declare axioms implicitly in Frame-style class definitions:

    ```python
    class ParentOfOne(Thing):
        equivalent_to = ObjectExactCardinality(1, HasChild, Person)
    ```

    UML:

    ```mermaid
    classDiagram
    AnonymousClassExpression <|-- ObjectExactCardinality
    ObjectExactCardinality --> "1" int : n
    ObjectExactCardinality --> "1" ObjectPropertyExpression : ope
    ObjectExactCardinality --> "1" ClassExpression : bce
    ```

    """

    n: int
    ope: ObjectPropertyExpression
    bce: ClassExpression

    # TODO: Aggregates



@dataclass
class AnonymousDataRange(ABC):
    """
    An base class for anonymous data ranges.
    """

    pass

@dataclass
class DatatypeDefinition:
    """
    A datatype definition.
    """

    dt: Datatype
    dr: DataRange

@dataclass
class DataIntersectionOf(AnonymousDataRange):
    """
    A DataRange that is the conjunction of DataRanges.
    """

    operands: Tuple[DataRange, ...] = field(default_factory=tuple)

    def __init__(self, *operands: DataRange):
        self.operands = operands

@dataclass
class DataUnionOf(AnonymousDataRange):
    """
    A DataRange that is the disjunction of DataRanges.
    """

    operands: Tuple[DataRange, ...] = field(default_factory=tuple)

    def __init__(self, *operands: DataRange):
        self.operands = operands

@dataclass
class DataComplementOf(AnonymousDataRange):
    """
    A DataRange that is the complement of a DataRange.
    """

    first: DataRange

@dataclass
class DataOneOf(AnonymousDataRange):
    """
    A DataRange that is the set of literals
    """

    operands: Tuple[Literal, ...] = field(default_factory=tuple)

    def __init__(self, *operands: Literal):
        self.operands = operands

@dataclass
class DataSomeValuesFrom(AnonymousDataRange):
    """
    A DataRange representing an existential restriction.
    """

    dpe: DataPropertyExpression
    dr: DataRange

@dataclass
class ObjectPropertyDomain(Axiom):
    """
    An object-property-domain axiom.

    Example of explicit declaration:

    ```python
    class HasParent(TopObjectProperty):
        '''A property that relates a person to their parent'''

    class Person(Thing):
        pass

    __axioms__ = [ ObjectPropertyDomain(HasParent, Person) ]
    ```

    Normally, the more Frame-style equivalent is preferred:

    ```python
    class HasParent(TopObjectProperty):
        '''A property that relates a person to their parent'''
        domain = Person
    ```

    UML:

    ```mermaid
    classDiagram
    Axiom <|-- ObjectPropertyDomain
    ObjectPropertyDomain --> "1" ObjectPropertyExpression : ope
    ObjectPropertyDomain --> "1" ClassExpression : ce
    ``
    """

    ope: ObjectPropertyExpression
    ce: ClassExpression

    def as_fol(self) -> Optional[Sentence]:
        return Forall([I, J], Implies(instance_of_op(I, J, self.ope), instance_of(I, self.ce)))


@dataclass
class DataPropertyDomain(Axiom):
    """
    A data-property-domain axiom.

    Example of explicit declaration:

    ```python
    class HasAge(TopDataProperty):
        '''A property that relates a person to their age'''

    class Person(Thing):
        pass

    __axioms__ = [ DataPropertyDomain(HasAge, int) ]
    ```

    Normally, the more Frame-style equivalent is preferred:

    ```python
    class HasAge(TopDataProperty):
        '''A property that relates a person to their age'''
        domain = Person
    ```

    UML:

    ```mermaid
    classDiagram
    Axiom <|-- DataPropertyDomain
    DataPropertyDomain --> "1" DataPropertyExpression : dpe
    DataPropertyDomain --> "1" ClassExpression : ce
    ``
    """

    dpe: DataPropertyExpression
    ce: ClassExpression

    def as_fol(self) -> Optional[Sentence]:
        return Forall([I, J], Implies(instance_of_dp(I, J, self.dpe), instance_of(I, self.ce)))

@dataclass
class ObjectPropertyRange(Axiom):
    """
    An object-property-range axiom.

    Example of explicit declaration:

    ```python
    class HasParent(TopObjectProperty):
        '''A property that relates a person to their parent'''

    class Person(Thing):
        pass

    __axioms__ = [ ObjectPropertyRange(HasParent, Person) ]
    ```

    Normally, the more Frame-style equivalent is preferred:

    ```python
    class HasParent(TopObjectProperty):
        '''A property that relates a person to their parent'''
        range = Person
    ```

    UML:

    ```mermaid
    classDiagram
    Axiom <|-- ObjectPropertyRange
    ObjectPropertyRange --> "1" ObjectPropertyExpression : ope
    ObjectPropertyRange --> "1" ClassExpression : ce
    ``
    """

    ope: ObjectPropertyExpression
    ce: ClassExpression

    def as_fol(self) -> Optional[Sentence]:
        return Forall([I, J], Implies(instance_of_op(I, J, self.ope), instance_of(J, self.ce)))

@dataclass
class DataPropertyRange(Axiom):
    """
    A data-property-range axiom.

    Example of explicit declaration:

    ```python
    class HasAge(TopDataProperty):
        '''A property that relates a person to their age'''

    class Person(Thing):
        pass

    __axioms__ = [ DataPropertyRange(HasAge, int) ]
    ```

    Normally, the more Frame-style equivalent is preferred:

    ```python
    class HasAge(TopDataProperty):
        '''A property that relates a person to their age'''
        range = int
    ```

    UML:

    ```mermaid
    classDiagram
    Axiom <|-- DataPropertyRange
    DataPropertyRange --> "1" DataPropertyExpression : dpe
    DataPropertyRange --> "1" DataRange : dr
    ``
    """

    dpe: DataPropertyExpression
    dr: DataRange

    # TODO
    #def as_fol(self) -> Optional[Sentence]:
    #    return Forall([I, J], Implies(instance_of_dp(I, J, self.dpe), instance_of(J, self.dr)))


@dataclass(frozen=True)
class Ontology(Fact):
    """
    An ontology.

    Note that an explicit declaration of an ontology is optional, as an ontology
    is inferred to be present in any python modules that contain declarations.

    Example:
    -------
    ```python
    class Animal(Thing):
        pass

    class Cat(Animal):
        pass

    class MyOntology(Ontology):
        iri = "http://example.org/ontology"
        axioms = [SubClassOf(Cat, Animal)]
    ```

    """

    axioms: ClassVar[Optional[List["Axiom"]]] = None

    @classmethod
    def to_sentences(cls) -> List[Sentence]:
        return [a for a in [axiom.as_fol() for axiom in cls.axioms or []] if a is not None]


OWL_AXIOM_REGISTRY = []
def owl_axioms(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    OWL_AXIOM_REGISTRY.append(func)
    return wrapper



from dataclasses import dataclass
from typing import ClassVar, Optional, Union

import yaml

from typedlogic import Sentence, Theory
from typedlogic.compiler import Compiler, ModelSyntax
from typedlogic.datamodel import as_object
from typedlogic.parsers.yaml_parser import YAMLParser


@dataclass
class YAMLCompiler(Compiler):

    default_suffix: ClassVar[str] = "yaml"
    parser_class = YAMLParser

    def compile(self, theory: Theory, syntax: Optional[Union[str, ModelSyntax]] = None, **kwargs) -> str:
        """
        Compile a Theory object into YAML code.

        Example:
        -------
            >>> from typedlogic import *
            >>> theory = Theory()
            >>> x = Variable("x")
            >>> theory.predicate_definitions = [PredicateDefinition("P", {"x": "str"}),
            ...              PredicateDefinition("Q", {"x": "str"})]
            >>> s = Implies(Term("P", x), Term("Q", x))
            >>> theory.add(sentence=s)
            >>> compiler = YAMLCompiler()
            >>> print(compiler.compile(theory))
            type: Theory
            constants: {}
            type_definitions: {}
            predicate_definitions:
            - type: PredicateDefinition
              predicate: P
              arguments:
                x: str
            - type: PredicateDefinition
              predicate: Q
              arguments:
                x: str
            sentence_groups:
            - type: SentenceGroup
              name: Sentences
              sentences:
              - type: Implies
                arguments:
                - type: Term
                  arguments:
                  - P
                  - type: Variable
                    arguments:
                    - x
                - type: Term
                  arguments:
                  - Q
                  - type: Variable
                    arguments:
                    - x
            ground_terms: []

        :param theory:
        :param syntax:
        :param kwargs:
        :return:

        """
        obj = as_object(theory)
        return yaml.dump(obj, sort_keys=False)

    def compile_sentence(self, sentence: Sentence, syntax: Optional[Union[str, ModelSyntax]] = None, **kwargs) -> str:
        obj = as_object(sentence)
        return yaml.dump(obj, sort_keys=False)



import logging
import sqlite3
from dataclasses import dataclass, field
from typing import ClassVar, Dict, Iterator, List, Optional, Union

import snakelog.common as snakelog
import snakelog.litelog as litelog
from snakelog.common import Var
from snakelog.souffle import SouffleSolver

import typedlogic as tlog
import typedlogic.pybridge
from typedlogic import FactMixin
from typedlogic.datamodel import (
    NotInProfileError,
    PredicateDefinition,
    Sentence,
    Term,
)
from typedlogic.parsers.pyparser.python_ast_utils import logger
from typedlogic.profiles import (
    ClassicDatalog,
    ExcludedProfile,
    MixedProfile,
    Profile,
    PropositionalLogic,
    UnsortedLogic,
)
from typedlogic.pybridge import fact_arg_values, fact_predicate
from typedlogic.solver import Method, Model, Solution, Solver
from typedlogic.transformations import to_horn_rules

logger = logging.getLogger(__name__)

PRIMITIVE_TYPES = (int, float, str, bool, bytes)

@dataclass
class SnakeLogSolver(Solver):
    """
    A solver that uses snakelog.

    Snakelog is a lightweight Datalog engine that uses SQLite as a backend,
    for more details see [this blog post](https://www.philipzucker.com/snakelog-post/).

    While Snakelog is only supports a limited subset of Datalog, it has the advantage of being
    **fast** and requiring no additional dependencies. It is well suited for simple traversal-style
    logic programming problems, such as the one below.

        >>> from typedlogic.integrations.frameworks.pydantic import FactBaseModel
        >>> from typedlogic import Implies, And, Variable
        >>> class AncestorOf(FactBaseModel):
        ...     ancestor: str
        ...     descendant: str
        >>> class ParentOf(FactBaseModel):
        ...     parent: str
        ...     child: str
        >>> solver = SnakeLogSolver(strict=True)
         >>> from typedlogic import SentenceGroup, PredicateDefinition
        >>> solver.add_predicate_definition(PredicateDefinition(predicate="AncestorOf", arguments={'ancestor': str, 'descendant': str}))
        >>> solver.add_predicate_definition(PredicateDefinition(predicate="ParentOf", arguments={'parent': str, 'child': str}))
        >>> solver.add_fact(ParentOf(parent='p1', child='p1a'))
        >>> solver.add_fact(ParentOf(parent='p1a', child='p1aa'))
        >>> X = Variable("X")
        >>> Y = Variable("Y")
        >>> Z = Variable("Z")
        >>> solver.add_sentence(Implies(Term(ParentOf.__name__, X, Y), Term(AncestorOf.__name__, X, Y)))
        >>> solver.add_sentence(Implies(And(Term(AncestorOf.__name__, X, Z),
        ...                                 Term(AncestorOf.__name__, Z, Y)),
        ...                             Term(AncestorOf.__name__, X, Y)))
        >>> model = solver.model()
        >>> facts = [str(f) for f in model.ground_terms]
        >>> for f in sorted(facts):
        ...     print(f)
        AncestorOf(p1, p1a)
        AncestorOf(p1, p1aa)
        AncestorOf(p1a, p1aa)
        ParentOf(p1, p1a)
        ParentOf(p1a, p1aa)

    This solver does not implement the open-world assumption.

        >>> from typedlogic.profiles import OpenWorld
        >>> solver.profile.impl(OpenWorld)
        False
    """

    _wrapped_solver: Optional[snakelog.BaseSolver] = None
    predicate_map: Optional[Dict[str, PredicateDefinition]] = None
    sentences: List[Sentence] = field(default_factory=list)

    profile: ClassVar[Profile] = MixedProfile(ClassicDatalog(), UnsortedLogic(), ExcludedProfile(PropositionalLogic()))

    methods_supported: ClassVar[List[Method]] = [
        Method(name="litelog", impl_class=litelog.Solver, is_default=True),
        Method(name="souffle", impl_class=SouffleSolver),
    ]

    @property
    def wrapped_solver(self) -> snakelog.BaseSolver:
        if self._wrapped_solver is None:
            impl_class = self.method.impl_class
            if impl_class is None:
                raise ValueError("No implementation class defined")
            self._wrapped_solver = impl_class()
        return self._wrapped_solver

    def check(self) -> Solution:
        return Solution(satisfiable=None)

    def models(self) -> Iterator[Model]:
        s = self.wrapped_solver
        s.run()

        facts = []
        if not self.predicate_map:
            raise ValueError("Predicates have not been defined")
        for p, pd in self.predicate_map.items():
            tbl = self.to_predicate(p)
            try:
                res = s.con.execute(f"SELECT * FROM {tbl}")
                for fact in res.fetchall():
                    bindings = dict(zip(pd.arguments.keys(), fact[0:], strict=False))
                    fact = Term(p, bindings)
                    facts.append(fact)
            except sqlite3.OperationalError:
                # TODO: better way to detect zero implications
                pass
        m = Model(
            source_object=s,
            ground_terms=facts
        )
        yield m

    def prove(self, sentence: Sentence) -> Optional[bool]:
        return super().prove(sentence)

    def add_fact(self, fact: FactMixin) -> None:
        p = self.to_predicate(fact_predicate(fact))
        atom = litelog.Atom(p, list(fact_arg_values(fact)))
        self.wrapped_solver.add(atom)
        self.sentences.append(fact)

    def add_sentence(self, sentence: Sentence) -> None:
        try:
            for sentence in to_horn_rules(sentence):
                for snakelog_expr in self.to_clauses(sentence):
                    self.wrapped_solver.add(snakelog_expr)
                    self.sentences.append(sentence)
        except NotInProfileError as e:
            logger.info(f"SKIPPING: {sentence} // {e}")
            if self.strict:
                raise e

    def _string_type(self) -> str:
        # TODO: remove after the following is fixed
        # https://github.com/philzook58/snakelog/issues/4
        if self.method_name == "souffle":
            return "symbol"
        return "TEXT"

    def add_predicate_definition(self, predicate_definition: PredicateDefinition) -> None:
        if not self.predicate_map:
            self.predicate_map = {}
        s = self.wrapped_solver
        string_type = self._string_type()
        arg_types = [string_type for _ in predicate_definition.arguments.keys()]
        sig = [self.to_predicate(predicate_definition.predicate)] + arg_types
        s.Relation(*sig)
        self.predicate_map[predicate_definition.predicate] = predicate_definition

    def to_predicate(self, predicate: str) -> str:
        return predicate.lower()

    def to_clauses(self, sentence: Sentence) -> List[Union[litelog.Clause, litelog.Atom]]:
        if isinstance(sentence, tlog.Forall):
            return self.to_clauses(sentence.sentence)
        if isinstance(sentence, tlog.Implied):
            return self.to_clauses(tlog.Implies(sentence.operands[1], sentence.operands[0]))
        if isinstance(sentence, tlog.Iff):
            return self.to_clauses(tlog.And(tlog.Implies(sentence.left, sentence.right),
                                            tlog.Implies(sentence.right, sentence.left)))
        if isinstance(sentence, tlog.And):
            sentences = []
            for s in sentence.operands:
                sentences.extend(self.to_clauses(s))
            return sentences
        return [self.to_clause(sentence)]

    def to_clause(self, sentence: Sentence) -> Union[litelog.Clause, litelog.Atom]:
        if isinstance(sentence, tlog.Forall):
            return self.to_clause(sentence.sentence)
        if isinstance(sentence, tlog.Implies):
            head = self.to_atom(sentence.consequent)
            body = self.to_body(sentence.antecedent)
            return litelog.Clause(head, body)
        if isinstance(sentence, tlog.Term):
            # unit clause
            return self.to_atom(sentence)
        raise NotInProfileError(f"Unknown clause type {type(sentence)} :: {sentence}")

    def to_atom(self, sentence: Sentence) -> litelog.Atom:
        if isinstance(sentence, tlog.Term):
            def _render_arg(arg):
                if arg is None:
                    return None
                if isinstance(arg, tlog.Variable):
                    # TODO: this should be the norm after normalization
                    arg = arg.name
                    return Var(arg.upper())
                else:
                    if isinstance(arg, PRIMITIVE_TYPES):
                        return arg
                    else:
                        return str(arg)
            return litelog.Atom(self.to_predicate(sentence.predicate), [_render_arg(a) for a in sentence.bindings.values()])
        if isinstance(sentence, typedlogic.pybridge.FactMixin):
            def _render_arg(arg):
                if arg is None:
                    return None
                return Var(arg.upper())
            p = self.to_predicate(fact_predicate(sentence))
            return litelog.Atom(p, [_render_arg(a) for a in fact_arg_values(sentence)])
        raise NotInProfileError(f"Unknown atom type {type(sentence)} :: {sentence}")

    def to_body(self, sentence: Sentence) -> litelog.Body:
        if isinstance(sentence, tlog.And):
            atoms = [self.to_atom(s) for s in sentence.operands]
            return litelog.Body(atoms)
        if isinstance(sentence, (tlog.Term, typedlogic.pybridge.FactMixin)):
            return litelog.Body([self.to_atom(sentence)])
        raise NotInProfileError(f"Unknown body type {type(sentence)} :: {sentence}")



    def dump(self) -> str:
        return str(self.wrapped_solver)




import hashlib
import random
import timeit
from dataclasses import dataclass
from typing import Optional, List, Any, Type, Callable, Dict, Union, Iterator


from typedlogic import Theory, Sentence, Term, NegationAsFailure, Variable
from typedlogic.datamodel import Extension
from typedlogic.integrations.solvers.clingo import ClingoSolver
from typedlogic.solver import Solver
from typedlogic.transformations import transform_sentence


@dataclass
class Benchmark:
    theory: Optional[Theory] = None
    ground_terms: Optional[List[Sentence]] = None
    goals: Optional[List[Sentence]] = None
    entities: Optional[List[Any]] = None
    entity_mapping: Optional[Dict[Any, Any]] = None


@dataclass
class BenchmarkSeed:
    theory: Theory
    ground_terms: Optional[List[Sentence]] = None
    candidate_goals: Optional[List[Sentence]] = None


@dataclass
class BenchmarkResult:
    benchmark: Benchmark
    solver_class_name: str
    score: float
    elapsed: float



def benchmark_from_seed(seed: BenchmarkSeed, benchmark_solver_class: Type[Solver] = ClingoSolver, num_positive=10, num_negative=10) -> Benchmark:
    benchmark = Benchmark(
        theory=seed.theory,
        ground_terms=seed.ground_terms,
    )
    if not seed.candidate_goals:
        raise ValueError("No candidate goals provided")
    candidate_goals = [t.to_model_object() if isinstance(t, Extension) else t for t in seed.candidate_goals]
    predicates = {t.predicate for t in candidate_goals if isinstance(t, Term)}
    if not predicates:
        raise ValueError(f"No predicates from {candidate_goals}")
    solver = benchmark_solver_class()
    solver.add(seed.theory)
    if not seed.ground_terms:
        raise ValueError("No ground terms provided")
    for t in seed.ground_terms:
        solver.add(t)
    #print(solver.base_theory)
    model = solver.model()
    if not model.ground_terms:
        raise ValueError(f"Solver {benchmark_solver_class} model has no ground terms for {predicates}")
    provable = []
    for p in predicates:
        #print(f"Querying: {p}")
        for t in model.iter_retrieve(p):
            provable.append(t)
    if not provable:
        raise ValueError(f"Solver {benchmark_solver_class} unable to find solutions for {predicates}")
    # take a random subsample of provable goals
    if len(provable) < num_positive:
        raise ValueError(f"Num provable = {len(provable)} < {num_positive}")
    goals: List[Sentence] = random.sample(provable, num_positive)
    random.shuffle(candidate_goals)
    n = 0
    for cg in candidate_goals:
        if cg not in provable:
            goals.append(NegationAsFailure(cg))
            n += 1
            if n >= num_negative:
                break
    random.shuffle(goals)
    benchmark.goals = goals
    return benchmark


def run_benchmark(benchmark: Benchmark, solver_class: Union[Solver, Type[Solver]]) -> BenchmarkResult:
    if isinstance(solver_class, Solver):
        solver: Solver = solver_class
        solver_class = type(solver)
    else:
        solver = solver_class()
    if not isinstance(solver, Solver):
        raise ValueError(f"Unexpected: {solver}")
    if not benchmark.theory:
        raise ValueError
    solver.add(benchmark.theory)
    for f in benchmark.ground_terms or []:
        solver.add(f.to_model_object() if isinstance(f, Extension) else f)
    start_time = timeit.default_timer()
    tps = 0
    goal_pairs = []
    goal_truths = {}
    if not benchmark.goals:
        raise ValueError
    for g in benchmark.goals or []:
        negative = False
        if isinstance(g, NegationAsFailure):
            negative = True
            g = g.negated
        goal_pairs.append((g, negative))
        goal_truths[str(repr(g))] = negative
    results = solver.prove_multiple([gp[0] for gp in goal_pairs])
    for g, is_proven in results:
        #[(_, negative)] = [gp for gp in goal_pairs if gp[0] == g]
        negative = goal_truths[str(repr(g))]
        correct = is_proven != negative
        if correct:
            tps += 1
    end_time = timeit.default_timer()
    total_time = end_time - start_time
    return BenchmarkResult(
        benchmark=benchmark,
        solver_class_name=solver_class.__name__,
        score=tps / len(benchmark.goals),
        elapsed=total_time,
    )


def run_benchmark_matrix(benchmarks: List[Benchmark], solver_classes: List[Union[Solver, Type[Solver]]]) -> Iterator[BenchmarkResult]:
    """
    Run a matrix of benchmarks against a matrix of solvers.

    :param benchmarks:
    :param solver_class:
    :return:
    """
    for benchmark in benchmarks:
        for solver_class in solver_classes:
            yield run_benchmark(benchmark, solver_class)


def create_random_id_function(prefix: str, max_id=10000) -> Callable[[], str]:
    """
    Create a function that generates random ids.

    :param prefix:
    :param max_id:
    :return:
    """
    unallocated = set(range(max_id))
    def assign(*args):
        nonlocal unallocated
        next_id = random.choice(list(unallocated))
        unallocated.remove(next_id)
        return f"{prefix}{next_id}"
    return assign

def randomize_entity_names(benchmark: Benchmark, mapping_func: Optional[Callable] = None):
    if mapping_func is None:
        mapping_func = lambda x: hashlib.md5(str(x).encode('utf-8')).hexdigest()
    emap = {}
    if not benchmark.entities:
        raise ValueError
    for e in benchmark.entities:
        emap[e] = mapping_func(e)
    def rewire_term(s: Sentence) -> Sentence:
        if isinstance(s, Extension):
            s = s.to_model_object()
        if isinstance(s, Term):
            return Term(s.predicate, *[emap[a] if not isinstance(a, Variable) and a in emap else a for a in s.values])
        return s
    benchmark.ground_terms = [transform_sentence(t, rewire_term) for t in benchmark.ground_terms or []]
    benchmark.goals = [transform_sentence(t, rewire_term) for t in benchmark.goals or []]
    benchmark.entity_mapping = emap


